# -*- coding: utf-8 -*-
# @Time    : 2022/7/7 11:01 上午
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : prettyPrint.py
import json
from typing import List
from treelib import Tree
from prettytable import PrettyTable

from ultipa.printer.treeNode import TreeNode
from ultipa.types.types import DataItem, ExplainPlan, Attr, ResultType, Path, Node, Edge
from ultipa.utils.convert import convertTableToDict

class PrettyPrint:
	count = 34
	tree = Tree()

	def printLine(self, name):
		num = int(PrettyPrint.count - len(name) / 2)
		print('-' * num + f'{name}' + '-' * num)

	def prettyStatus(self, data):
		y = PrettyTable(['Code', 'Message'], title="STATUS", end=10, align="l")
		y.add_row([data.code, data.message])
		print(y)

	def prettyStatistics(self, data):
		y = PrettyTable(['edgeAffected', 'nodeAffected', 'engineCost', 'totalCost'], title="Statistics", end=10,
						align="l")
		y.add_row([data.edgeAffected, data.nodeAffected, data.engineCost, data.totalCost])
		print(y)

	def prettyDataList(self, data: List[dict]):
		propertList = []
		if len(data) > 0:
			if len(data[0].keys()) > 0:
				propertList.extend(list(data[0].keys()))
			y = PrettyTable(propertList, title=f"Alias: Data", align="l")
			for d in data:
				li = []
				for p in propertList:
					li.append(d.get(p))
				y.add_row(li)
			print(y)
		else:
			return None

	def prettyData(self, data: dict):
		propertList = []
		if len(data.keys()) > 0:
			propertList.extend(list(data.keys()))
		y = PrettyTable(propertList, title=f"Alias: Data", align="l")
		li = []
		for p in propertList:
			li.append(data.get(p))
		y.add_row(li)
		print(y)

	def prettyNode(self, data: DataItem):
		schemaDict = {}
		for node in data.asNodes():
			propertList = ["id", "uuid", "schema"]
			if not schemaDict.get(node.getSchema()):
				if node.values:
					newPropertList = propertList + list(node.values.keys())
					schemaDict.update({node.getSchema(): newPropertList})
				else:
					schemaDict.update({node.getSchema(): propertList})

		for key in schemaDict:
			y = PrettyTable(schemaDict.get(key), title=f"Alias: {data.alias} AliasType: {data.type} Schema: {key}",
							align="l")
			for newNode in data.asNodes():
				if newNode.getSchema() == key:
					li = [newNode.getID(), newNode.getUUID(), newNode.getSchema()]
					if newNode.values:
						li.extend(list(newNode.values.values()))
					y.add_row(li)
				else:
					continue
			print(y)

	def prettyAttrNode(self, alias, nodes: List[Node]):
		ret = []
		schemaDict = {}
		for node in nodes:
			propertList = ["id", "uuid", "schema"]
			if not schemaDict.get(node.getSchema()):
				if node.values:
					newPropertList = propertList + list(node.values.keys())
					schemaDict.update({node.getSchema(): newPropertList})
				else:
					schemaDict.update({node.getSchema(): propertList})

		for key in schemaDict:
			y = PrettyTable(schemaDict.get(key), title=f"Alias: {alias} AliasType: NODE Schema: {key}",
							align="l")
			for newNode in nodes:
				if newNode.getSchema() == key:
					li = [newNode.getID(), newNode.getUUID(), newNode.getSchema()]
					if newNode.values:
						li.extend(list(newNode.values.values()))
					y.add_row(li)
				else:
					continue
			ret.append(y)
		return ret

	def prettyEdge(self, data: DataItem):
		schemaDict = {}
		for edge in data.asEdges():
			propertList = ["uuid", "from_uuid", "to_uuid", "from_id", "to_id", "schema"]
			if not schemaDict.get(edge.getSchema()):
				if edge.values:
					newPropertList = propertList + list(edge.values.keys())
					schemaDict.update({edge.getSchema(): newPropertList})
				else:
					schemaDict.update({edge.getSchema(): propertList})

		for key in schemaDict:
			y = PrettyTable(schemaDict.get(key), title=f"Alias: {data.alias} AliasType: {data.type} Schema: {key}",
							align="l")
			for newEdge in data.asEdges():
				if newEdge.getSchema() == key:
					li = [newEdge.getUUID(), newEdge.getFromUUID(), newEdge.getToUUID(), newEdge.getFrom(),
						  newEdge.getTo(), newEdge.getSchema()]
					if newEdge.values:
						li.extend(list(newEdge.values.values()))
					y.add_row(li)
				else:
					continue
			print(y)

	def prettyAttrEdge(self, alias, edges: List[Edge]):
		ret = []
		schemaDict = {}
		for edge in edges:
			propertList = ["uuid", "from_uuid", "to_uuid", "from_id", "to_id", "schema"]
			if not schemaDict.get(edge.getSchema()):
				if edge.values:
					newPropertList = propertList + list(edge.values.keys())
					schemaDict.update({edge.getSchema(): newPropertList})
				else:
					schemaDict.update({edge.getSchema(): propertList})

		for key in schemaDict:
			y = PrettyTable(schemaDict.get(key), title=f"Alias: {alias} AliasType: EDGE Schema: {key}", align="l")
			for newEdge in edges:
				if newEdge.getSchema() == key:
					li = [newEdge.getUUID(), newEdge.getFromUUID(), newEdge.getToUUID(), newEdge.getFrom(),
						  newEdge.getTo(), newEdge.getSchema()]
					if newEdge.values:
						li.extend(list(newEdge.values.values()))
					y.add_row(li)
				else:
					continue
			ret.append(y)
		return ret

	def prettyTable(self, data: DataItem):

		if data.alias == "_algoList":
			headers = data.asTable().getHeaders()
			rows = data.asTable().getRows()
			table_rows_dict = convertTableToDict(rows, headers)
			algo_list = []
			for data1 in table_rows_dict:
				algo_list.append(json.loads(data1.get("param")))
			if len(algo_list) > 0:
				propertList = algo_list[0].keys()
				y = PrettyTable(propertList, title=f"Alias: {data.alias} AliasType: {data.type}", align="l")
				for algo in algo_list:
					row = []
					for key in propertList:
						row.append(algo.get(key))
					y.add_row(row)
			else:
				return
		else:
			propertList = []
			if len(data.asTable().rows) > 0:
				for index, header in enumerate(data.asTable().getHeaders()):
					propertList.append(header.get("property_name"))
			else:
				return

			y = PrettyTable(propertList, title=f"Alias: {data.alias} AliasType: {data.type}", align="l")
			for row in data.asTable().getRows():
				y.add_row(row)
		print(y)

	def _PrintPath(self, path: Path):
		pathStr = ""
		end = ""
		for index, edge in enumerate(path.edges):
			d1 = "-"
			d2 = "-"
			node = path.getNodes()[index]
			if edge.getFrom() == node.getID():
				d2 = "->"
				if index == len(path.edges) - 1:
					end = f" ({str(edge.getTo())})"
			else:
				d1 = "<-"
				if index == len(path.edges) - 1:
					end = f" ({str(edge.getFrom())})"

			pathStr += f"({node.getID()}) {d1} [{edge.getUUID()}] {d2}"
			if end:
				pathStr += end
		return pathStr

	def prettyPath(self, data: DataItem):
		propertList = ["#", "Path"]
		y = PrettyTable(propertList, title=f"Alias: {data.alias} AliasType: {data.type}", align="l")
		for i, path in enumerate(data.asPaths()):
			li = [i]
			pathStr = self._PrintPath(path)
			li.append(pathStr)
			y.add_row(li)
		print(y)

	def _formatAttrListData(self, data: list, alias: str):
		if isinstance(data,list) and len(data) > 0:
			if isinstance(data[0], Path):
				li = []
				for i in data:
					li.append(self._PrintPath(i))
				return str(li)
			elif isinstance(data[0], Node):
				return self.prettyAttrNode(alias, data)

			elif isinstance(data[0], Edge):
				return self.prettyAttrEdge(alias, data)
			else:
				return str(data)
		else:
			return str(data)

	def prettyAttr(self, data: DataItem):
		if data.asAttr().type in [ResultType.RESULT_TYPE_PATH, ResultType.RESULT_TYPE_EDGE,
								  ResultType.RESULT_TYPE_NODE]:
			propertList = ["#", data.alias]
			y = PrettyTable(propertList, title=f"Alias: {data.alias} AliasType: {data.type}", align="l")
			for i, adata in enumerate(data.asAttr().values):
				ret = self._formatAttrListData(adata, data.alias)
				y.add_row([i + 1, ret])
			print(y)
		else:
			propertList = ["#", data.alias]
			y = PrettyTable(propertList, title=f"Alias: {data.alias} AliasType: {data.type}", align="l")
			if len(data.asAttr().values) > 0:
				for i, adata in enumerate(data.asAttr().values):
					ret = self._formatAttrListData(adata, data.alias)
					y.add_row([i + 1, ret])
			else:
				y.add_row([1, str(data.asAttr().values)])
			print(y)

	def prettyArray(self, data: DataItem):
		propertList = [data.alias]
		y = PrettyTable(propertList, title=f"Alias: {data.alias} AliasType: {data.type}", align="l")
		for array in data.asArray().elements:
			y.add_row([json.dumps(array, ensure_ascii=False)])
		print(y)

	def prettyTree(self, plans: List[ExplainPlan]):
		def buildTree(plans: List[ExplainPlan]):
			if plans == None or len(plans) == 0:
				return TreeNode()
			tree = TreeNode(plans[0])
			del (plans[0])
			for i in range(tree.explain.children_num):
				node = buildTree(plans)
				tree.childNodes.append(node)
			return tree

		root = buildTree(plans)
		self._printTree(root, root.explain.id, True)
		ret = self.tree.show(stdout=False)
		propertList = ["data"]
		y = PrettyTable(propertList, title=f"Alias: tree AliasType: ExplainPlan", align="l")
		y.add_row([ret])
		print(y)

	def _printTree(self, root: TreeNode, parter, isRoot=False):
		if isRoot:
			self.tree.create_node(root.explain.uql, parter)
		else:
			self.tree.create_node(root.explain.uql, identifier=root.explain.id, parent=parter)
		if len(root.childNodes) > 0:
			parter = root.explain.id
			for child in root.childNodes:
				self._printTree(child, parter)
		else:
			return



# -*- coding: utf-8 -*-
# @Time    : 2023/7/18 09:15
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : __init__.py.py
from ultipa.types import ULTIPA, ULTIPA_REQUEST,ULTIPA_RESPONSE
from ultipa.types.types import UltipaConfig
from ultipa.types.types_request import *
from ultipa.utils.ufilter import ufilter as FILTER
from ultipa.connection.connection import Connection




from ultipa.types.types import TruncateType
from ultipa.operate.base_extra import BaseExtra
from ultipa.utils import UQLMAKER, CommandList
from ultipa.types import ULTIPA_REQUEST, ULTIPA, ULTIPA_RESPONSE

JSONSTRING_KEYS = ["graph_privileges", "system_privileges", "policies", "policy", "privilege"]
formatdata = ['graph_privileges']


class TruncateExtra(BaseExtra):

	def truncate(self, request: ULTIPA_REQUEST.Truncate,
				 requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.ResponseCommon:
		command = CommandList.truncate
		requestConfig.graphName = request.graphSetName
		uqlMaker = UQLMAKER(command, commonParams=requestConfig)
		uqlMaker.addParam("graph", request.graphSetName)
		if request.dbType:
			if request.dbType == TruncateType.NODES:
				if request.all:
					uqlMaker.addParam("nodes", "*")
				if not request.all and request.schema:
					uqlMaker.addParam("nodes", "@" + request.schema, notQuotes=True)
			if request.dbType == TruncateType.EDGES:
				if request.all:
					uqlMaker.addParam("edges", "*")
				if not request.all and request.schema:
					uqlMaker.addParam("edges", "@" + request.schema, notQuotes=True)

		# if request.all and not request.dbType:
		#     uqlMaker = UQLMAKER(command,commandP=request.dbType,commonParams=requestConfig)

		return self.UqlUpdateSimple(uqlMaker)

	def compact(self, request: ULTIPA_REQUEST.Graph,
				requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.ResponseCommon:
		command = CommandList.compact
		uqlMaker = UQLMAKER(command, commonParams=requestConfig)
		uqlMaker.addParam("graph", request.graph)
		return self.UqlUpdateSimple(uqlMaker)

	def mount(self, request: ULTIPA_REQUEST.Mount,
			  requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.ResponseCommon:
		commonP = []
		if request.graph:
			commonP = request.graph
			requestConfig.graphName = request.graph
		uqlMaker = UQLMAKER(command=CommandList.mount, commonParams=requestConfig)
		uqlMaker.setCommandParams(commandP=commonP)
		return self.UqlUpdateSimple(uqlMaker)

	def unmount(self, request: ULTIPA_REQUEST.Unmount,
				requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.ResponseCommon:
		commonP = []
		if request.graph:
			commonP = request.graph
			requestConfig.graphName = request.graph
		uqlMaker = UQLMAKER(command=CommandList.unmount, commonParams=requestConfig)
		uqlMaker.setCommandParams(commandP=commonP)
		return self.UqlUpdateSimple(uqlMaker)



import json

from ultipa.operate.base_extra import BaseExtra
from ultipa.types import ULTIPA, ULTIPA_REQUEST, ULTIPA_RESPONSE
from ultipa.utils import UQLMAKER, CommandList

# BOOL_KEYS = ["index","lte"]
from ultipa.utils.ResposeFormat import ResponseKeyFormat
from ultipa.utils.propertyUtils import getPropertyTypesDesc

BOOL_KEYS = ["index", "lte"]
REPLACE_KEYS = {
	"name": "propertyName",
	"type": "propertyType",
}


class PropertyExtra(BaseExtra):

	def listProperty(self,requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.ResponseListProperty:
		return self.showProperty(requestConfig)

	def showProperty(self,requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.ResponseListProperty:
		command = CommandList.showProperty
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		res = self.UqlListSimple(uqlMaker=uqlMaker,responseKeyFormat=ResponseKeyFormat(keyReplace=REPLACE_KEYS, boolKeys=BOOL_KEYS),
								 isSingleOne=False)
		return res

	def getProperty(self, request: ULTIPA_REQUEST.GetProperty,
					requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.ResponseGetProperty:
		'''
		:EXP: conn.listProperty(ULTIPA_REQUEST.GetProperty(type=ULTIPA.DBType.DBNODE))
		:param request: ULTIPA_REQUEST
		:return:
		'''
		if request:
			if request.type == ULTIPA.DBType.DBNODE:
				command = CommandList.showNodeProperty
			elif request.type == ULTIPA.DBType.DBEDGE:
				command = CommandList.showEdgeProperty
			else:
				command = CommandList.showNodeProperty
			if request.schemaName:
				commandp = ['@' + request.schemaName]
			else:
				commandp = ''
		else:
			command = CommandList.showNodeProperty
			commandp = ''

		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		uqlMaker.setCommandParams(commandP=commandp)
		res = self.UqlListSimple(uqlMaker=uqlMaker,responseKeyFormat=ResponseKeyFormat(keyReplace=REPLACE_KEYS, boolKeys=BOOL_KEYS),
								 isSingleOne=False)
		if res.data!=None and len(res.data) > 0:
			res.data = res.data[0].data
		return res


	def createProperty(self, request: ULTIPA_REQUEST.CreateProperty,
					   requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		:EXP: conn.createProperty(ULTIPA_REQUEST.CreatProperty(type=ULTIPA.DBType.DBNODE, name='test_int',property_type=ULTIPA.CreatePropertyType.PROPERTY_INT))
		:param request: ULTIPA_REQUEST
		:return:
		'''
		command = request.type == ULTIPA.DBType.DBNODE and CommandList.createNodeProperty or CommandList.createEdgeProperty
		commandP = [request.schemaName.schemaName, request.schemaName.propertyName, getPropertyTypesDesc(request.propertyType,request.subTypes)]

		if request.description:
			commandP.append(request.description)
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		uqlMaker.setCommandParams(commandP=commandP)
		res = self.uqlSingle(uqlMaker)
		return res

	def dropProperty(self, request: ULTIPA_REQUEST.DropProperty,
					 requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		:EXP: conn.dropProperty(ULTIPA_REQUEST.DropProperty(type=ULTIPA.DBType.DBNODE, name='test'))
		:param request: ULTIPA_REQUEST
		:return:
		'''
		command = request.type == ULTIPA.DBType.DBNODE and CommandList.dropNodeProperty or CommandList.dropEdgeProperty
		commandP = request.schemaName.toString
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		uqlMaker.setCommandParams(commandP=commandP)
		res = self.uqlSingle(uqlMaker)
		return res

	def alterProperty(self, request: ULTIPA_REQUEST.AlterProperty,
					  requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		:EXP: conn.alterProperty(ULTIPA_REQUEST.DropProperty(type=ULTIPA.DBType.DBNODE, name='test'))
		:param request: ULTIPA_REQUEST
		:return:
		'''
		command = request.type == ULTIPA.DBType.DBNODE and CommandList.alterNodeProperty or CommandList.alterEdgeProperty
		commandP = request.schemaName.toString
		update_dict = {}
		if request.new_name:
			update_dict.setdefault('name', request.new_name)
		if request.description:
			update_dict.update({'description': request.description})
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		uqlMaker.setCommandParams(commandP)
		uqlMaker.addParam("set", update_dict)
		res = self.uqlSingle(uqlMaker)
		return res



import json

from ultipa.operate.base_extra import BaseExtra
from ultipa.types import ULTIPA_REQUEST, ULTIPA, ULTIPA_RESPONSE
from ultipa.utils import UQLMAKER, CommandList
from ultipa.utils.ResposeFormat import ResponseKeyFormat

REPLACE_KEYS = {
	"graph": "name",
}


class GraphExtra(BaseExtra):

	def listGraph(self,
				  requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.ResponseListGraph:
		return self.showGraph(requestConfig)

	def showGraph(self,
				  requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.ResponseListGraph:
		uqlMaker = UQLMAKER(command=CommandList.showGraph, commonParams=requestConfig)
		uqlMaker.setCommandParams("")
		res = self.UqlListSimple(uqlMaker=uqlMaker, responseKeyFormat=ResponseKeyFormat(keyReplace=REPLACE_KEYS))
		return res

	def getGraph(self, request: ULTIPA_REQUEST.Graph,
				 requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.ResponseGraph:
		uqlMaker = UQLMAKER(command=CommandList.showGraph, commonParams=requestConfig)
		uqlMaker.setCommandParams(request.graph)

		res = self.UqlListSimple(uqlMaker=uqlMaker, responseKeyFormat=ResponseKeyFormat(keyReplace=REPLACE_KEYS))
		if res.status.code == ULTIPA.Code.SUCCESS and res.data != None and len(res.data) > 0:
			res.data = res.data[0]
		return res

	def createGraph(self, request: ULTIPA_REQUEST.Graph,
					requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		uqlMaker = UQLMAKER(command=CommandList.createGraph, commonParams=requestConfig)
		if request.description:
			uqlMaker.setCommandParams([request.graph, request.description])
		else:
			uqlMaker.setCommandParams(request.graph)
		res = self.uqlSingle(uqlMaker)
		return res

	def dropGraph(self, request: ULTIPA_REQUEST.Graph,
				  requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		uqlMaker = UQLMAKER(command=CommandList.dropGraph, commonParams=requestConfig)
		uqlMaker.setCommandParams(request.graph)
		res = self.uqlSingle(uqlMaker)
		return res

	def alterGraph(self, request: ULTIPA_REQUEST.AlterGraph,
				   requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		requestConfig.graphName = request.oldGraphName
		uqlMaker = UQLMAKER(command=CommandList.alterGraph, commonParams=requestConfig)
		uqlMaker.setCommandParams(request.oldGraphName)
		data = {"name": request.newGraphName}
		if request.newDescription is not None:
			data.update({'description': request.newDescription})
		uqlMaker.addParam("set", data)
		res = self.uqlSingle(uqlMaker)
		return res



from typing import List

from ultipa import Connection, UltipaConfig

from ultipa.connection.connectionPoolMaker import ConnectionPool
from ultipa.connection.connectionBase import ConnectionBase


class UltipaConnectionPool():
	def __init__(self, defaultConfig: UltipaConfig = UltipaConfig()):
		self.pool = ConnectionPool(lambda: Connection.NewConnection(defaultConfig=defaultConfig), max_size=10)

	def get_conn(self) -> ConnectionBase:
		with self.pool.item() as ultipa_cli:
			return ultipa_cli

	def __del__(self):
		conn = self.get_conn()
		wrapped = self.pool._wrapper(conn)
		self.pool._destroy(wrapped)

	def destroyConn(self, conn):
		wrapped = self.pool._wrapper(conn)
		self.pool._destroy(wrapped)



# -*- coding: utf-8 -*-
# @Time    : 2023/7/18 09:20
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : connectionBase.py

import logging
import time
import schedule
from ultipa import ParameterException
from ultipa.connection.clientInfo import ClientInfo
from ultipa.connection.clientType import ClientType
from ultipa.connection.hostManagerControl import HostManagerControl, RAFT_GLOBAL
from ultipa.connection.uqlHelper import UQLHelper
from ultipa.proto import ultipa_pb2
from ultipa.types.types import UltipaConfig
from ultipa.types.types_response import PropertyTable
from ultipa.utils import UQLMAKER, CommandList
from ultipa.utils.ResposeFormat import ResponseKeyFormat
from ultipa.utils.common import GETLEADER_TIMEOUT
from ultipa.utils.convert import convertTableToDict, convertToListAnyObject
from ultipa.utils.format import FormatType
from ultipa.utils.logger import LoggerConfig
from ultipa.types import ULTIPA, ULTIPA_REQUEST, ULTIPA_RESPONSE
from ultipa.utils.ultipaSchedule import run_continuously

class ConnectionBase:
	hostManagerControl: HostManagerControl
	username: str
	password: str
	crtPath: str
	defaultConfig: UltipaConfig
	runSchedule: object = None

	def __init__(self, host: str, defaultConfig: UltipaConfig, crtFilePath: str = None):
		self.host = host
		self.username = defaultConfig.username
		self.password = defaultConfig.password
		self.crtPath = crtFilePath
		self.defaultConfig = defaultConfig
		self.crt = None
		if crtFilePath:
			try:
				with open(f'{crtFilePath}', 'rb') as f:
					self.crt = f.read()
			except Exception as e:
				raise ParameterException(err=e)
		self.hostManagerControl = HostManagerControl(self.host, self.username, self.password,
													 self.defaultConfig.maxRecvSize, self.crt,
													 consistency=defaultConfig.consistency)

		self.defaultConfig.defaultGraph = defaultConfig.defaultGraph or "default"
		self.defaultConfig.timeoutWithSeconds = defaultConfig.timeoutWithSeconds or 15
		self.defaultConfig.responseWithRequestInfo = defaultConfig.responseWithRequestInfo or False
		self.defaultConfig.consistency = defaultConfig.consistency
		self.graphSetName = self.defaultConfig.defaultGraph
		self.count = 0

		if not self.defaultConfig.uqlLoggerConfig and self.defaultConfig.Debug:
			self.defaultConfig.uqlLoggerConfig = LoggerConfig(name="ultipa", fileName="",
															  isStream=self.defaultConfig.Debug, isWriteToFile=False,
															  level=logging.INFO)

	def getGraphSetName(self, currentGraphName: str, uql: str = "", isGlobal: bool = False):
		if isGlobal:
			return RAFT_GLOBAL
		if uql:
			parse = UQLHelper(uql)
			if parse.uqlIsGlobal():
				return RAFT_GLOBAL
			c1 = parse.parseRet.getFirstCommands()
			c2 = f"{c1}().{parse.parseRet.getSecondCommands()}"
			if c2 in [CommandList.mount, CommandList.unmount, CommandList.truncate]:
				graphName = parse.parseRet.getCommandsParam(1)
				if graphName:
					return graphName
		return currentGraphName or self.defaultConfig.defaultGraph

	def getTimeout(self, timeout: int):
		return timeout or self.defaultConfig.timeoutWithSeconds

	def getClientInfo(self, clientType: int = ClientType.Default, graphSetName: str = '', uql: str = '',
					  isGlobal: bool = False, ignoreRaft: bool = False, useHost: str = None, useMaster: bool = False,
					  timezone=None, timeZoneOffset=None):
		goGraphName = self.getGraphSetName(currentGraphName=graphSetName, uql=uql, isGlobal=isGlobal)
		if not ignoreRaft and not self.hostManagerControl.getHostManger(goGraphName).raftReady:
			refreshRet = self.refreshRaftLeader(self.hostManagerControl.initHost,
												ULTIPA_REQUEST.RequestConfig(graphName=goGraphName))
			self.hostManagerControl.getHostManger(goGraphName).raftReady = refreshRet

		clientInfo = self.hostManagerControl.chooseClientInfo(type=clientType, uql=uql, graphSetName=goGraphName,
															  useHost=useHost, useMaster=useMaster)
		metadata = clientInfo.getMetadata(goGraphName, timezone, timeZoneOffset)
		return ClientInfo(Rpcsclient=clientInfo.Rpcsclient, Controlsclient=clientInfo.Controlsclient, metadata=metadata,
						  graphSetName=goGraphName, host=clientInfo.host)



	def getRaftLeader(self, requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()):
		resRaftLeader = self.__autoGetRaftLeader(host=self.host, requestConfig=requestConfig)
		RaftStatus = FormatType.getRaftStatus(resRaftLeader)
		return ULTIPA_RESPONSE.Response(RaftStatus)

	def __getRaftLeader(self, requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()):
		if requestConfig == None:
			graphSetName = None
		else:
			if not requestConfig.graphName:
				graphSetName = 'default'
			else:
				graphSetName = requestConfig.graphName

		clientInfo = self.getClientInfo(clientType=ClientType.Leader, graphSetName=graphSetName, ignoreRaft=True,
										useMaster=requestConfig.useMaster)
		res = clientInfo.Controlsclient.GetLeader(ultipa_pb2.GetLeaderRequest(), metadata=clientInfo.metadata,
												  timeout=GETLEADER_TIMEOUT)

		return FormatType.Response(_res=res, host=self.host)

	def __autoGetRaftLeader(self, host: str, requestConfig: ULTIPA_REQUEST.RequestConfig, retry=0):
		'''内部用，所以返回值自定义'''
		conn = ConnectionBase(host=host, crtFilePath=self.crtPath, defaultConfig=self.defaultConfig)
		try:
			res = conn.__getRaftLeader(requestConfig)
		except Exception as e:
			self.hostManagerControl.initHost = ""
			if host in self.defaultConfig.hosts:
				self.defaultConfig.hosts.remove(host)
			return {
				"code": ULTIPA.Code.FAILED,
				"message": str(e._state.code) + ' : ' + str(e._state.details)
			}
		status = res.status
		if status.code == ULTIPA.Code.SUCCESS:
			# status.clusterInfo.raftPeers.remove(host)  # 不要主节点
			self.hostManagerControl.initHost = host
			for i in status.clusterInfo.raftPeers:
				if i.host == host:
					status.clusterInfo.raftPeers.remove(i)
			return {
				"code": status.code,
				"message": status.message,
				'leaderHost': host,
				"followersPeerInfos": list(filter(lambda x: x != host, status.clusterInfo.raftPeers)),
				"leaderInfos": status.clusterInfo.leader,
			}
		elif status.code == ULTIPA.Code.NOT_RAFT_MODE:
			return {
				"code": status.code,
				"message": status.message,
				"leaderHost": host,
				"followersPeerInfos": [],
				"leaderInfos": status.clusterInfo.leader
			}
		elif status.code in [ULTIPA.Code.RAFT_REDIRECT, ULTIPA.Code.RAFT_LEADER_NOT_YET_ELECTED,
							 ULTIPA.Code.RAFT_NO_AVAILABLE_FOLLOWERS, ULTIPA.Code.RAFT_NO_AVAILABLE_ALGO_SERVERS]:
			if retry > 2:
				return {
					"code": status.code,
					"message": status.message,
					"redirectHost": res.status.clusterInfo.redirect
				}
			if status.code != ULTIPA.Code.RAFT_REDIRECT:
				time.sleep(0.3)
			if status.code == ULTIPA.Code.RAFT_REDIRECT:
				host = res.status.clusterInfo.redirect
			return self.__autoGetRaftLeader(host=host, requestConfig=requestConfig, retry=retry + 1)

		return {
			"code": status.code,
			"message": status.message
		}

	def refreshRaftLeader(self, redirectHost: str, requestConfig: ULTIPA_REQUEST.RequestConfig):
		# hosts = [redirectHost] if redirectHost else self.hostManagerControl.getAllHosts()
		hosts = [redirectHost] if redirectHost else []
		goGraphName = self.getGraphSetName(requestConfig.graphName)

		for h in self.defaultConfig.hosts:
			if h not in hosts:
				hosts.append(h)
		for host in hosts:
			resRaftLeader = self.__autoGetRaftLeader(host=host, requestConfig=requestConfig)
			code = resRaftLeader["code"]
			if code == ULTIPA.Code.SUCCESS:
				leaderHost = resRaftLeader["leaderHost"]
				followersPeerInfos = resRaftLeader["followersPeerInfos"]
				leaderInfos = resRaftLeader["leaderInfos"]
				hostManager = self.hostManagerControl.upsetHostManger(goGraphName, leaderHost)
				hostManager.setClients(leaderHost=leaderHost, followersPeerInfos=followersPeerInfos,
									   leaderInfos=leaderInfos)
				return True
		# elif code == ULTIPA.Code.RAFT_REDIRECT:
		# 	return False
		return False

	def stopConnectionAlive(self):
		if self.runSchedule != None:
			self.runSchedule.set()

	def keepConnectionAlive(self, timeIntervalSeconds: int = None):
		timeIntervalSeconds = self.defaultConfig.heartBeat if timeIntervalSeconds == None else timeIntervalSeconds

		def test_allconn():
			goGraphName = self.defaultConfig.defaultGraph
			for host in self.hostManagerControl.getAllClientInfos(goGraphName):
				res = host.Controlsclient.SayHello(ultipa_pb2.HelloUltipaRequest(name="test"),
												   metadata=host.getMetadata(goGraphName, None, None))
				# print(host.host,res.message)
				if self.defaultConfig.uqlLoggerConfig is None:
					self.defaultConfig.uqlLoggerConfig = LoggerConfig(name="HeartBeat", fileName=None,
																	  isWriteToFile=False,
																	  isStream=True)
				self.defaultConfig.uqlLoggerConfig.getlogger().info(f"HeartBeat:{host.host}--{res.message}")

		schedule.every().second.do(test_allconn)
		self.runSchedule = run_continuously(timeIntervalSeconds)




import json
import re
from typing import List
from ultipa.utils import errors
from ultipa.utils.ufilter.new_ufilter import Filter
from ultipa.utils.ufilter.ufilter import FilterBase

class CommandList:
    ab = "ab"
    khop = "khop"
    nodes = "find().nodes"
    edges = "find().edges"
    deleteNodes = "delete().nodes"
    deleteEdges = "delete().edges"
    updateNodes = "update().nodes"
    updateEdges = "update().edges"
    template = "t"
    autoNet = "autoNet"
    autoNetByPart = "autoNetByPart"
    nodeSpread = "spread"
    insert = "insert"
    upsert = "upsert"
    showProperty = "show().property"
    showNodeProperty = "show().node_property"
    showEdgeProperty = "show().edge_property"
    createNodeProperty = "create().node_property"
    createEdgeProperty = "create().edge_property"
    dropNodeProperty = "drop().node_property"
    dropEdgeProperty = "drop().edge_property"
    alterNodeProperty = "alter().node_property"
    alterEdgeProperty = "alter().edge_property"
    showSchema = "show().schema"
    showNodeSchema = "show().node_schema"
    showEdgeSchema = "show().edge_schema"
    createNodeSchema = "create().node_schema"
    createEdgeSchema = "create().edge_schema"
    alterNodeSchema = "alter().node_schema"
    alterEdgeSchema = "alter().edge_schema"
    dropNodeSchema = "drop().node_schema"
    dropEdgeSchema = "drop().edge_schema"
    lteNode = "LTE().node_property"
    lteEdge = "LTE().edge_property"
    ufeNode = "UFE().node_property"
    ufeEdge = "UFE().edge_property"
    createNodeIndex = "create().node_index"
    createEdgeIndex = "create().edge_index"
    createNodeFulltext = "create().node_fulltext"
    createEdgeFulltext = "create().edge_fulltext"
    showIndex = "show().index"
    showNodeIndex = "show().node_index"
    showEdgeIndex = "show().edge_index"
    showFulltext = "show().fulltext"
    showNodeFulltext = "show().node_fulltext"
    showEdgeFulltext = "show().edge_fulltext"
    dropNodeIndex = "drop().node_index"
    dropEdgeIndex = "drop().edge_index"
    dropNodeFulltext = "drop().node_fulltext"
    dropEdgeFulltext = "drop().edge_fulltext"
    stat = "stats"
    algo = "algo"
    algo_dv = "algo_dv"
    showPrivilege = "show().privilege"
    grant = "grant().user"
    revoke = "revoke().user"
    showAlgo = "show().algo"
    showGraph = "show().graph"
    createGraph = "create().graph"
    dropGraph = "drop().graph"
    alterGraph = "alter().graph"
    truncate = "truncate"
    compact = "compact"
    showUser = "show().user"
    getUser = "show().user"
    getSelfInfo = "show().self"
    createUser = "create().user"
    alterUser = "alter().user"
    dropUser = "drop().user"
    createPolicy = "create().policy"
    alterPolicy = "alter().policy"
    dropPolicy = "drop().policy"
    showPolicy = "show().policy"
    getPolicy = "show().policy"
    showTask = "show().task"
    clearTask = "clear().task"
    pauseTask = "pause().task"
    resumeTask = "resume().task"
    stopTask = "stop().task"
    top = "top"
    kill = "kill"
    mount = "mount().graph"
    unmount = "unmount().graph"

def _replace(value:str):
    template = re.compile(r"\'?\"?(point\([^)]*\))\'?\"?")
    matches = re.search(template, value)
    if matches:
        return re.sub(r"\'?\"?(point\([^)]*\))\'?\"?",matches.group(1),value)
    return value

class UQLMAKER:
    def __init__(self, command: CommandList, commandP: any = None, commonParams=None):
        self._command = command
        self._commandP = commandP
        self.commonParams = commonParams
        self._params: List[object] = []
        self.templateParams: List[object] = []

    @staticmethod
    def PointFunction(latitude: float, longitude: float):
        return "point({latitude:%s, longitude:%s})" % (latitude,longitude)

    @staticmethod
    def PointString(latitude: float, longitude: float):
        return f"POINT({latitude} {longitude})"

    def setCommandParams(self, commandP: any):
        if commandP:
            if not isinstance(commandP, list):
                self._commandP = [commandP]
            else:
                self._commandP = commandP
            newcommandP = []
            for comm in self._commandP:
                if isinstance(comm, list) or isinstance(comm, dict) or isinstance(comm, int):
                    newcommandP.append(str(comm))
                    continue
                if isinstance(comm, Filter):
                    if isinstance(comm, str):
                        continue
                    newcommandP.append("{%s}" % (comm.builder()))
                    continue
                if not comm:
                    newcommandP.append(comm)
                    continue
                if comm.startswith('{'):
                    newcommandP.append(comm)
                    continue
                if comm.startswith('@'):
                    newcommandP.append(comm)
                else:
                    newcommandP.append(json.dumps(comm, ensure_ascii=False))

            commandP = ','.join(newcommandP)
        else:
            return
        if type(commandP) == object:
            commandP = json.dumps(commandP, ensure_ascii=False)
        if isinstance(commandP, Filter):
            commandP = commandP.builder()
        self._commandP = commandP

    def addTemplateParams(self, templateParams: List[object]):
        self.templateParams = templateParams

    def addParam(self, key: str, value: any, required: bool = True, notQuotes=False):
        try:
            if notQuotes:
                self._params.append({"key": key, "value": value})
                return
            _notStringify = False
            if type(value) == bool:
                if value:
                    required = False
                value = ""
            if required:
                if isinstance(value, list) or value or value == 0:
                    pass
                else:
                    return
            if "filter" == key:
                self.addParam("node_filter", value)
                self.addParam("edge_filter", value)
                return

            if key in ["filter", "node_filter", "edge_filter"]:
                _notStringify = True
                if isinstance(value, FilterBase):
                    value = value.builder()

            if key == 'return':
                if type(value) == list:
                    value = ','.join([i.toString for i in value])
                elif type(value) == str:
                    self._params.append({"key": key, "value": value})
                    return
                else:
                    value = value.toString
                self._params.append({"key": key, "value": value})
                return
            if key in ['into', 'as']:
                self._params.append({"key": key, "value": value})
                return

            if key in ['graph_privileges']:
                if isinstance(value, list):
                    value = [{v.toDict().get('name'): v.toDict().get('values')} for v in value]
                else:
                    value = [{value.toDict().get('name'): value.toDict().get('values')}]

            if type(value) == object or type(value) == dict or (
                    type(value) == str and len(value) > 0 and not _notStringify):
                value = json.dumps(value,ensure_ascii=False)
            if isinstance(value,list):
                value =[i for i in value]

            if value == [] or value == None:
                return
            self._params.append({"key": key, "value": value})
        except Exception as e:
            raise errors.ParameterException(e)

    def toString(self):
        uql = ""
        str_return = ""
        if self._commandP:
            self._commandP = self._commandP
            uql += "{}({})".format(self._command, self._commandP)
        else:
            uql += "{}({})".format(self._command, '')
        if len(self.templateParams) > 0:
            for tp in self.templateParams:
                filterString = ''
                if tp.filter:
                    if isinstance(tp.filter, FilterBase):
                        tp.filter = tp.filter.builder()
                    filterString = tp.filter
                node_filter_str = ''
                if tp.__dict__.get('node_fitler'):
                    if isinstance(tp.node_fitler, FilterBase):
                        tp.node_fitler = tp.node_fitler.builder()
                    node_filter_str = f'.nf{tp.node_fitler}'
                step = tp.__dict__.get('steps')
                stepStr = ''
                if step:
                    stepStr = f'[{":".join(step)}]'
                uql += f'.{tp.name}({tp.alias or ""}{filterString or ""}){node_filter_str}{stepStr}'
        if len(self._params) > 0:
            ps = []
            for p in self._params:
                value = p["value"]
                if p['key'] in ["return", "limit", "as"]:
                    fstr = "{} {}".format(p['key'], value)
                    str_return += fstr + " "
                    continue

                if p['key'] in ["select_node_properties", "select_edge_properties", "srcs",
                                "dests", "return"]:
                    fstr = "{}({})".format(p["key"], value)
                else:
                    fstr = "{}({})".format(p["key"], value)
                fstr = _replace(fstr)
                ps.append(fstr)
            if len(ps) > 0:
                uql += "." + ".".join(ps)
        uql += " " + str_return
        return uql.strip()


class UQLParmas():
    def __init__(self):
        self.commands = []
        self.commandParam = {}
        self.params = {}
        self.paramsOriginal = {}

    def getFirstCommands(self):
        if len(self.commands)>0:
            return self.commands[0]
        return None

    def getSecondCommands(self):
        if len(self.commands)>1:
            return self.commands[1]
        return None

    def getCommands(self,index):
        if len(self.commands)>index:
            return self.commands[index]
        return None

    def getCommandsParam(self,index):
        if len(self.commands) > index:
            return self.commandParam.get(self.commands[index])
        return None



class UQL:

    @staticmethod
    def uqlObjectExample(uqlStr):
        ret = UQLParmas()
        ret.uql = uqlStr
        return ret

    @staticmethod
    def parse(uqlStr):
        commandReg = '([a-zA-Z]*)\(([^\(|^\)]*)\)'
        matchAll = re.findall(commandReg, uqlStr)
        result = UQL.uqlObjectExample(uqlStr)
        for i,m in enumerate(matchAll):
            name, value = m
            result.commands.append(name)
            if isinstance(value,str):
                value = value.replace("\"","").replace("'","")
            result.commandParam.update({name:value})
        return result

    @staticmethod
    def parse_globle(uqlStr):
        reg = '(.*\(\)\.[A-Za-z]+\(?)'
        matchAll = re.findall(reg, uqlStr)
        result = UQL.uqlObjectExample(uqlStr)
        index = 0
        if matchAll:
            for value in matchAll:
                result.commands.append(value.strip("("))
                index += 1
            return result




from ultipa.operate.base_extra import BaseExtra
from ultipa.types import ULTIPA_REQUEST, ULTIPA, ULTIPA_RESPONSE
from ultipa.utils import UQLMAKER, CommandList


class LteUfeExtra(BaseExtra):

	def lte(self, request: ULTIPA_REQUEST.LTE,
			requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		:param requestConfig:
		:EXP: conn.lte(ULTIPA_REQUEST.LteUfe(property='name',type=ULTIPA.DBType.DBNODE))
		:param request: ULTIPA_REQUEST
		:return:
		'''
		command = request.type == ULTIPA.DBType.DBNODE and CommandList.lteNode or CommandList.lteEdge
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		uqlMaker.setCommandParams(request.schemaName.toString)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res

	def ufe(self, request: ULTIPA_REQUEST.UFE,
			requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		:param requestConfig:
		:EXP: conn.ufe(ULTIPA_REQUEST.LteUfe(property='name',type=ULTIPA.DBType.DBEDGE))
		:param request: ULTIPA_REQUEST
		:return:
		'''
		command = request.type == ULTIPA.DBType.DBNODE and CommandList.ufeNode or CommandList.ufeEdge
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		uqlMaker.setCommandParams(request.schemaName.toString)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res



from ultipa.operate.base_extra import BaseExtra
from ultipa.types import ULTIPA_REQUEST, ULTIPA_RESPONSE, ULTIPA
from ultipa.utils import UQLMAKER, CommandList
from ultipa.utils.errors import ParameterException
from ultipa.utils.ResposeFormat import ResponseKeyFormat

JSONSTRING_KEYS = ["graphPrivileges", "systemPrivileges", "policies", "policy", "privilege"]
formatdata = ['graph_privileges']


class PolicyExtra(BaseExtra):

	def showPolicy(self,
				   requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.ResponseListPolicy:
		'''
		:EXP: conn.listPolicy()
		:return:
		'''
		uqlMaker = UQLMAKER(command=CommandList.showPolicy, commonParams=requestConfig)
		res = self.UqlListSimple(uqlMaker=uqlMaker,
								 responseKeyFormat=ResponseKeyFormat(jsonKeys=JSONSTRING_KEYS, dataFormat=formatdata))
		return res

	def showPrivilege(self,
					  requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.ResponseListPrivilege:
		'''
		:EXP: conn.listPolicy()
		:return:
		'''
		uqlMaker = UQLMAKER(command=CommandList.showPrivilege, commonParams=requestConfig)
		res = self.UqlListSimple(uqlMaker=uqlMaker,
								 responseKeyFormat=ResponseKeyFormat(jsonKeys=JSONSTRING_KEYS))
		return res

	def getPolicy(self, request: ULTIPA_REQUEST.GetPolicy,
				  requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.ResponsePolicy:
		'''
		:param requestConfig:
		:EXP: conn.getPolicy(ULTIPA_REQUEST.Policy(name='sales',privileges=['QUERY']))
		:param request: ULTIPA_REQUEST
		:return:
		'''
		uqlMaker = UQLMAKER(command=CommandList.getPolicy, commonParams=requestConfig)
		uqlMaker.setCommandParams(request.name)
		res = self.UqlListSimple(uqlMaker=uqlMaker,
								 responseKeyFormat=ResponseKeyFormat(jsonKeys=JSONSTRING_KEYS, dataFormat=formatdata))
		if isinstance(res.data, list) and len(res.data) > 0:
			res.data = res.data[0]
		return res

	def createPolicy(self, request: ULTIPA_REQUEST.CreatePolicy,
					 requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		:param requestConfig:
		:EXP: conn.createPolicy(ULTIPA_REQUEST.Policy(name='sales',privileges=['QUERY']))
		:param request: ULTIPA_REQUEST
		:return:
		'''
		uqlMaker = UQLMAKER(command=CommandList.createPolicy, commonParams=requestConfig)
		paramsP = [request.name]
		if request.graph_privileges:
			paramsP.append(request.graph_privileges)
		else:
			paramsP.append({})
		if request.system_privileges:
			paramsP.append(request.system_privileges)
		else:
			paramsP.append([])
		if request.policies:
			paramsP.append(request.policies)
		else:
			paramsP.append([])
		uqlMaker.setCommandParams(paramsP)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res

	def alterPolicy(self, request: ULTIPA_REQUEST.AlterPolicy,
					requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		:param requestConfig:
		:EXP: conn.alterPolicy(ULTIPA_REQUEST.AlterPolicy(name='sales', privileges=['QUERY','DELETE']))
		:param request: ULTIPA_REQUEST
		:return:
		'''
		uqlMaker = UQLMAKER(command=CommandList.alterPolicy, commonParams=requestConfig)
		uqlMaker.setCommandParams(request.name)
		params = {}
		if request.system_privileges is not None:
			params.update({"system_privileges": request.system_privileges})
		if request.graph_privileges is not None:
			params.update({"graph_privileges": request.graph_privileges})
		if request.policies is not None:
			params.update({"policies": request.policies})
		uqlMaker.addParam('set', params, notQuotes=True)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res

	def dropPolicy(self, request: ULTIPA_REQUEST.DropPolicy,
				   requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		:param requestConfig:
		:EXP: conn.dropPolicy(ULTIPA_REQUEST.DropPolicy(name='sales'))
		:param request: ULTIPA_REQUEST
		:return:
		'''
		uqlMaker = UQLMAKER(command=CommandList.dropPolicy, commonParams=requestConfig)
		uqlMaker.setCommandParams(request.name)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res

	def grantPolicy(self, request: ULTIPA_REQUEST.GrantPolicy,
					requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		:param requestConfig:
		:EXP: conn.grant(ULTIPA_REQUEST.GrantRevoke(username='sales', privileges=['QUERY']))
		:param request: ULTIPA_REQUEST
		:return:
		'''
		uqlMaker = UQLMAKER(command=CommandList.grant, commonParams=requestConfig)
		if request.username:
			uqlMaker.setCommandParams(request.username)
		else:
			raise ParameterException(err='username is a required parameter')

		paramsDict = {}
		if request.graph_privileges:
			paramsDict.setdefault('graph_privileges', request.graph_privileges)

		if request.system_privileges:
			paramsDict.setdefault('system_privileges', request.system_privileges)

		if request.policies:
			paramsDict.setdefault('policies', request.policies)
		uqlMaker.addParam('params', paramsDict, notQuotes=True)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res

	def revokePolicy(self, request: ULTIPA_REQUEST.RevokePolicy,
					 requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		:param requestConfig:
		:EXP: conn.revoke(ULTIPA_REQUEST.GrantRevoke(username='sales', privileges=['QUERY']))
		:param request: ULTIPA_REQUEST
		:return:
		'''
		uqlMaker = UQLMAKER(command=CommandList.revoke, commonParams=requestConfig)
		if request.username:
			uqlMaker.setCommandParams(request.username)
		else:
			raise ParameterException(err='username is a required parameter')

		paramsDict = {}
		if request.graph_privileges:
			paramsDict.setdefault('graph_privileges', request.graph_privileges)

		if request.system_privileges:
			paramsDict.setdefault('system_privileges', request.system_privileges)

		if request.policies:
			paramsDict.setdefault('policies', request.policies)
		uqlMaker.addParam('params', paramsDict, notQuotes=True)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res



'''
ultipa 常量设置
'''

LIMIT = 1

DEPTH = 1

GETLEADER_TIMEOUT = 5
# dateTime 类型是否转为时间格式的字符串  1970-01-01 08:00:00
DateTimeIs2Str = True
# Timestamp 类型是否转为时间格式的字符串  1970-01-01 08:00:00 （根据当前时区进行转换）
TimestampIs2Str = False
# 精度控制返回7位
Float_Accuracy = True


from typing import List
from ultipa.operate.base_extra import BaseExtra
from ultipa.proto import ultipa_pb2
from ultipa.types import ULTIPA_REQUEST, ULTIPA_RESPONSE, ULTIPA
from ultipa.utils.format import FormatType


class DownloadExtra(BaseExtra):

	def _download(self, request: ULTIPA_REQUEST.Download,
				 requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> List[
		ULTIPA_RESPONSE.Response]:
		'''
		:param requestConfig:
		:param task_id:
		:param file_name:
		:EXP:
			with open('./XXXXX.csv', 'ab+') as f:
				for data_flow in ret:
					data = data_flow.chunk
					f.write(data)
		:return: stream
		'''
		downResponse = ULTIPA_RESPONSE.Response()
		try:
			clientInfo = self.getClientInfo(graphSetName=requestConfig.graphName, useMaster=requestConfig.useMaster)
			res = clientInfo.Controlsclient.DownloadFileV2(
				ultipa_pb2.DownloadFileRequestV2(file_name=request.fileName, task_id=request.taskId),
				metadata=clientInfo.metadata)
			if request.savePath:
				with open(request.savePath, 'wb+') as f:
					for data_flow in res:
						status = FormatType.status(data_flow.status)
						downResponse.status = status
						if status.code != ULTIPA.Code.SUCCESS:
							yield downResponse
							break
						data = data_flow.chunk
						f.write(data)
				yield downResponse
			else:
				for data_flow in res:
					ultipa_response = ULTIPA_RESPONSE.Response()
					status = FormatType.status(data_flow.status)
					ultipa_response.status = status
					if status.code != ULTIPA.Code.SUCCESS:
						yield ultipa_response
						break
					ultipa_response.data = data_flow.chunk
					yield ultipa_response
		except Exception as e:
			downResponse.status = ULTIPA.Status(code=ULTIPA.Code.UNKNOW_ERROR, message=str(e._state.details))
			yield downResponse

	def download(self, request: ULTIPA_REQUEST.Download,
				 requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> List[
		ULTIPA_RESPONSE.Response]:
		return  list(self._download(request,requestConfig))


# -*- coding: utf-8 -*-
# @Time    : 2023/7/18 17:17
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : base_extra.py
import copy
import csv
import json
import types
from typing import List

from ultipa import ParameterException
from ultipa.connection.clientType import ClientType
from ultipa.connection.commonUql import GetPropertyBySchema
from ultipa.connection.connectionBase import ConnectionBase
from ultipa.connection.uqlHelper import UQLHelper
from ultipa.proto import ultipa_pb2
from ultipa.types import ULTIPA, ULTIPA_REQUEST, ULTIPA_RESPONSE
from ultipa.types.types_response import PropertyTable
from ultipa.utils import UQLMAKER, CommandList
from ultipa.utils.ResposeFormat import ResponseKeyFormat
from ultipa.utils.convert import convertTableToDict, convertToListAnyObject
from ultipa.utils.format import FormatType
from ultipa.utils.raftRetry import RetryHelp
from ultipa.utils.ultipa_datetime import getTimeZoneOffset, getTimeOffsetSeconds


class BaseExtra(ConnectionBase):

	def test(self,
			 requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.Response:
		testResponse = ULTIPA_RESPONSE.Response()
		returnReq = ULTIPA.ReturnReq(requestConfig.graphName, "test", None, None, False)
		try:
			clientInfo = self.getClientInfo(useHost=requestConfig.useHost, useMaster=requestConfig.useMaster)
			name = 'Test'
			res = clientInfo.Controlsclient.SayHello(ultipa_pb2.HelloUltipaRequest(name=name),
													 metadata=clientInfo.metadata)
			returnReq.host = clientInfo.host
			if (res.message == name + " Welcome To Ultipa!"):
				if self.defaultConfig.uqlLoggerConfig:
					self.defaultConfig.uqlLoggerConfig.getlogger().info(res.message)

				testResponse.status = ULTIPA.Status(code=res.status.error_code, message=res.status.msg)
			else:
				testResponse.status = ULTIPA.Status(code=res.status.error_code, message=res.status.msg)
		except Exception as e:
			testResponse = ULTIPA_RESPONSE.Response()
			try:
				message = str(e._state.details)
			except:
				message = str(e)
			testResponse.status = ULTIPA.Status(code=ULTIPA.Code.UNKNOW_ERROR, message=message)
		if self.defaultConfig.responseWithRequestInfo:
			testResponse.req = returnReq
		return testResponse

	def stats(self,
			  requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.ResponseStat:
		uqlMaker = UQLMAKER(command=CommandList.stat, commonParams=requestConfig)
		ret = self.UqlListSimple(uqlMaker=uqlMaker, isSingleOne=True)
		if ret.status.code == ULTIPA.Code.SUCCESS:
			ret.data = ret.data[0]
		return ret

	def exportData(self, request: ULTIPA_REQUEST.Export,
				   requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()):
		try:
			req = ultipa_pb2.ExportRequest(db_type=request.type, limit=request.limit,
										   select_properties=request.properties, schema=request.schema)

			clientInfo = self.getClientInfo(graphSetName=requestConfig.graphName, useMaster=requestConfig.useMaster)
			res = clientInfo.Controlsclient.Export(req, metadata=clientInfo.metadata)
			res = FormatType.exportResponse(_res=res, timeZone=requestConfig.timeZone,
											timeZoneOffset=requestConfig.timeZoneOffset)
			return res
		except Exception as e:
			errorRes = ULTIPA_RESPONSE.Response()
			try:
				message = str(e._state.code) + ' : ' + str(e._state.details)
			except:
				message = 'UNKNOW ERROR'

			errorRes.status = ULTIPA.Status(code=ULTIPA.Code.UNKNOW_ERROR, message=message)
			errorRes.req = ULTIPA.ReturnReq(self.graphSetName, "exportData",
											requestConfig.useHost if requestConfig.useHost else self.host,
											requestConfig.retry,
											False)
			return errorRes

	def uql(self, uql: str,
			requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		request = ultipa_pb2.UqlRequest()
		request.uql = uql
		request.timeout = self.getTimeout(requestConfig.timeoutWithSeconds)
		if requestConfig.threadNum is not None:
			request.thread_num = requestConfig.threadNum
		ultipaRes = ULTIPA_RESPONSE.UltipaResponse()
		uqlLoggerConfig = self.defaultConfig.uqlLoggerConfig
		if requestConfig.graphName == '' and self.defaultConfig.defaultGraph != '':
			requestConfig.graphName = self.defaultConfig.defaultGraph

		if self.defaultConfig.consistency != self.hostManagerControl.consistency:
			self.hostManagerControl.consistency = self.defaultConfig.consistency

		if self.defaultConfig.maxRecvSize != self.hostManagerControl.maxRecvSize:
			self.hostManagerControl.maxRecvSize = self.defaultConfig.maxRecvSize
		onRetry = copy.deepcopy(requestConfig.retry)
		while onRetry.current < onRetry.max:
			try:
				import pytz
				getTimeZoneOffset(requestConfig, self.defaultConfig)
				timeZone = requestConfig.timeZone if requestConfig.timeZone else self.defaultConfig.timeZone
				timeZoneOffset = requestConfig.timeZoneOffset if requestConfig.timeZoneOffset else self.defaultConfig.timeZoneOffset
				timeZoneOffset = getTimeOffsetSeconds(timeZoneOffset)
				clientInfo = self.getClientInfo(graphSetName=requestConfig.graphName, uql=uql,
												useHost=requestConfig.useHost,
												useMaster=requestConfig.useMaster, timezone=timeZone,
												timeZoneOffset=timeZoneOffset)
				if uqlLoggerConfig:
					uqlLoggerConfig.getlogger().info(
						f'Begin UQL: {uql} graphSetName: {clientInfo.graphSetName} Host: {clientInfo.host}')
				uqlIsExtra = UQLHelper.uqlIsExtra(uql)
				if uqlIsExtra:
					res = clientInfo.Controlsclient.UqlEx(request, metadata=clientInfo.metadata)
				else:
					res = clientInfo.Rpcsclient.Uql(request, metadata=clientInfo.metadata)

				if not requestConfig.stream:
					ultipaRes = FormatType.uqlMergeResponse(res, timeZone, timeZoneOffset)
				else:
					ultipaRes = FormatType.uqlResponse(res, timeZone, timeZoneOffset)

				if self.defaultConfig.responseWithRequestInfo and not requestConfig.stream:
					ultipaRes.req = ULTIPA.ReturnReq(clientInfo.graphSetName, uql, clientInfo.host, onRetry,
													 uqlIsExtra)
				if not isinstance(ultipaRes, types.GeneratorType) and RetryHelp.checkRes(ultipaRes):
					onRetry.current += 1
					continue
				else:
					return ultipaRes

			except Exception as e:
				onRetry.current += 1
				if uqlLoggerConfig:
					uqlLoggerConfig.getlogger().info(
						f'Begin Retry [{onRetry.current}]- clientInfo host: {clientInfo.host} graphSetName: {clientInfo.graphSetName}')
				self.hostManagerControl.getHostManger(requestConfig.graphName).raftReady = False
				try:
					message = str(e._state.code) + ' : ' + str(e._state.details)
				except:
					message = str(e)
				ultipaRes.status = ULTIPA.Status(code=ULTIPA.Code.UNKNOW_ERROR, message=message)
				ultipaRes.req = ULTIPA.ReturnReq(requestConfig.graphName, uql,
												 requestConfig.useHost if requestConfig.useHost else self.host,
												 onRetry, False)

		return ultipaRes

	def uqlSingle(self, uqlMaker: UQLMAKER) -> ULTIPA_RESPONSE.UltipaResponse:
		res = self.uql(uqlMaker.toString(), uqlMaker.commonParams)
		return res


	def UqlListSimple(self, uqlMaker: UQLMAKER, responseKeyFormat: ResponseKeyFormat = None,
					  isSingleOne: bool = True) -> ULTIPA_RESPONSE.Response:
		res = self.uqlSingle(uqlMaker)

		if res.status.code != ULTIPA.Code.SUCCESS:
			simplrRes = ULTIPA_RESPONSE.Response(res.status, res.items)
			return simplrRes

		if not isSingleOne:
			retList = []
			for alias in res.aliases:
				item = res.items.get(alias.alias)
				table = item.data
				table_rows = table.rows
				table_rows_dict = convertTableToDict(table_rows, table.headers)
				if responseKeyFormat:
					table_rows_dict = responseKeyFormat.changeKeyValue(table_rows_dict)
				data = convertToListAnyObject(table_rows_dict)
				retList.append(PropertyTable(name=table.name, data=data))
			simplrRes = ULTIPA_RESPONSE.Response(res.status, retList)
			simplrRes.req = res.req
			return simplrRes

		alisFirst = res.aliases[0].alias if len(res.aliases) > 0 else None
		firstItem = res.items.get(alisFirst)
		if firstItem:
			table_rows = firstItem.data.rows
			table_rows_dict = convertTableToDict(table_rows, firstItem.data.headers)
			if responseKeyFormat:
				table_rows_dict = responseKeyFormat.changeKeyValue(table_rows_dict)
			data = convertToListAnyObject(table_rows_dict)
			simplrRes = ULTIPA_RESPONSE.Response(res.status, data)
			simplrRes.req = res.req
			simplrRes.statistics = res.statistics
			return simplrRes
		else:
			return res

	def UqlUpdateSimple(self, uqlMaker: UQLMAKER):
		res = self.uqlSingle(uqlMaker)

		if res.status.code != ULTIPA.Code.SUCCESS:
			return ULTIPA_RESPONSE.Response(res.status, statistics=res.statistics)

		if res.req:
			ret = ULTIPA_RESPONSE.Response(res.status, statistics=res.statistics)
			ret.req = res.req
			return ret
		return ULTIPA_RESPONSE.Response(res.status, statistics=res.statistics)

	def insertNodesBatchBySchema(self, schema: ULTIPA_REQUEST.Schema, rows: List[ULTIPA.EntityRow],
								 config: ULTIPA_REQUEST.InsertConfig) -> ULTIPA_RESPONSE.InsertResponse:
		config.useMaster = True
		if config.graphName == '' and self.defaultConfig.defaultGraph != '':
			config.graphName = self.defaultConfig.defaultGraph

		clientInfo = self.getClientInfo(clientType=ClientType.Update, graphSetName=config.graphName,
										useMaster=config.useMaster)

		nodetable = FormatType.makeEntityNodeTable(schema, rows, getTimeZoneOffset(requestConfig=config,
																				   defaultConfig=self.defaultConfig))

		_nodeTable = ultipa_pb2.EntityTable(schemas=nodetable.schemas, entity_rows=nodetable.nodeRows)
		request = ultipa_pb2.InsertNodesRequest()
		request.silent = config.silent
		request.insert_type = config.insertType
		request.graph_name = config.graphName
		request.node_table.MergeFrom(_nodeTable)
		res = clientInfo.Rpcsclient.InsertNodes(request, metadata=clientInfo.metadata)

		status = FormatType.status(res.status)
		uqlres = ULTIPA_RESPONSE.Response(status=status)
		reTry = RetryHelp.check(self, config, uqlres)
		if reTry.canRetry:
			config.retry = reTry.nextRetry
			return self.insertNodesBatchBySchema(schema, rows, config)

		uRes = ULTIPA_RESPONSE.ResponseBulk()
		uRes.uuids = [i for i in res.uuids]
		errorDict = {}
		for i, data in enumerate(res.ignore_error_code):
			try:
				index = rows[res.ignore_indexes[i]]._getIndex()
			except Exception as e:
				try:
					index = res.ignore_indexes[i]
				except Exception as e:
					index = i
			if index is None:
				try:
					index = res.ignore_indexes[i]
				except Exception as e:
					index = i
			errorDict.update({index: data})
		uRes.errorItem = errorDict
		uqlres.data = uRes
		uqlres.total_time = res.time_cost
		uqlres.engine_time = res.engine_time_cost
		if self.defaultConfig.responseWithRequestInfo:
			uqlres.req = ULTIPA.ReturnReq(config.graphName, "InsertNodesBatchBySchema", clientInfo.host, reTry,
										  False)
		return uqlres

	def insertEdgesBatchBySchema(self, schema: ULTIPA_REQUEST.Schema, rows: List[ULTIPA.EntityRow],
								 config: ULTIPA_REQUEST.InsertConfig) -> ULTIPA_RESPONSE.InsertResponse:
		config.useMaster = True
		if config.graphName == '' and self.defaultConfig.defaultGraph != '':
			config.graphName = self.defaultConfig.defaultGraph

		clientInfo = self.getClientInfo(clientType=ClientType.Update, graphSetName=config.graphName,
										useMaster=config.useMaster)

		edgetable = FormatType.makeEntityEdgeTable(schema=schema, rows=rows,
												   timeZoneOffset=getTimeZoneOffset(requestConfig=config,
																					defaultConfig=self.defaultConfig))

		_edgeTable = ultipa_pb2.EntityTable(schemas=edgetable.schemas, entity_rows=edgetable.edgeRows)
		request = ultipa_pb2.InsertEdgesRequest()
		request.silent = config.silent
		request.insert_type = config.insertType
		request.graph_name = config.graphName
		request.create_node_if_not_exist = config.createNodeIfNotExist
		request.edge_table.MergeFrom(_edgeTable)
		res = clientInfo.Rpcsclient.InsertEdges(request, metadata=clientInfo.metadata)

		status = FormatType.status(res.status)
		uqlres = ULTIPA_RESPONSE.Response(status=status)
		reTry = RetryHelp.check(self, config, uqlres)
		if reTry.canRetry:
			config.retry = reTry.nextRetry
			return self.insertEdgesBatchBySchema(schema, rows, config)

		uRes = ULTIPA_RESPONSE.ResponseBulk()
		uRes.uuids = [i for i in res.uuids]
		errorDict = {}
		for i, data in enumerate(res.ignore_error_code):
			try:
				index = rows[res.ignore_indexes[i]]._getIndex()
			except Exception as e:
				try:
					index = res.ignore_indexes[i]
				except Exception as e:
					index = i
			if index is None:
				try:
					index = res.ignore_indexes[i]
				except Exception as e:
					index = i
			errorDict.update({index: data})
		uRes.errorItem = errorDict
		uqlres.data = uRes
		uqlres.total_time = res.time_cost
		uqlres.engine_time = res.engine_time_cost
		if self.defaultConfig.responseWithRequestInfo:
			uqlres.req = ULTIPA.ReturnReq(config.graphName, "InsertEdgesBatchBySchema", clientInfo.host, reTry,
										  False)
		return uqlres

	def insertNodesBatchAuto(self, nodes: List[ULTIPA.EntityRow],
							 config: ULTIPA_REQUEST.InsertConfig) -> ULTIPA_RESPONSE.ResponseBatchAutoInsert:
		Result = {}
		schemaDict = {}
		batches = {}
		schemaRet = self.uql(GetPropertyBySchema.node, config)
		if schemaRet.status.code == ULTIPA.Code.SUCCESS:
			for aliase in schemaRet.aliases:
				if aliase.alias == '_nodeSchema':
					schemaDict = convertTableToDict(schemaRet.alias(aliase.alias).data.rows,
													schemaRet.alias(aliase.alias).data.headers)
			if not schemaDict:
				raise ParameterException(err='Please create Node Schema.')
		else:
			raise ParameterException(err=schemaRet.status.message)
		for index, node in enumerate(nodes):
			node._index = index
			if batches.get(node.schema) is None:
				batches[node.schema] = ULTIPA_REQUEST.Batch()
				find = list(filter(lambda x: x.get('name') == node.schema, schemaDict))
				if find:
					findSchema = find[0]
					propertyList = FormatType.checkProperty(node, json.loads(findSchema.get("properties")))
					reqSchema = ULTIPA_REQUEST.Schema(node.schema, propertyList)
					batches[node.schema].Schema = reqSchema
				else:
					if node.schema is None:
						raise ParameterException(err=f"Row [{index}]:Please set schema name for node.")
					else:
						raise ParameterException(err=f"Row [{index}]:Node Schema not found {node.schema}.")

			batches.get(node.schema).Nodes.append(node)
		for key in batches:
			batch = batches.get(key)
			Result.update(
				{key: self.insertNodesBatchBySchema(schema=batch.Schema, rows=batch.Nodes, config=config)})

		newStatusMsg = ""
		newCode = None
		for i, key in enumerate(Result):
			ret = Result.get(key)
			newStatusMsg += f"{key}:{ret.status.message} "
			if ret.status.code != ULTIPA.Code.SUCCESS and not newCode:
				newCode = ret.status.code
		if newCode is None:
			newCode = ULTIPA.Code.SUCCESS
		status = ULTIPA_RESPONSE.Status(newCode, newStatusMsg)
		newResponse = ULTIPA_RESPONSE.ResponseBatchAutoInsert(status=status)
		newResponse.data = Result
		return newResponse

	def insertEdgesBatchAuto(self, edges: List[ULTIPA.EntityRow],
							 config: ULTIPA_REQUEST.InsertConfig) -> ULTIPA_RESPONSE.ResponseBatchAutoInsert:
		Result = {}
		schemaDict = []
		batches = {}
		schemaRet = self.uql(GetPropertyBySchema.edge, config)
		if schemaRet.status.code == ULTIPA.Code.SUCCESS:
			for aliase in schemaRet.aliases:
				if aliase.alias == '_edgeSchema':
					schemaDict = convertTableToDict(schemaRet.alias(aliase.alias).data.rows,
													schemaRet.alias(aliase.alias).data.headers)
			if not schemaDict:
				raise ParameterException(err='Please create Edge Schema.')
		else:
			raise ParameterException(err=schemaRet.status.message)
		for index, edge in enumerate(edges):
			edge._index = index
			if batches.get(edge.schema) == None:
				batches[edge.schema] = ULTIPA_REQUEST.Batch()
				find = list(filter(lambda x: x.get('name') == edge.schema, schemaDict))
				if find:
					findSchema = find[0]
					propertyList = FormatType.checkProperty(edge, json.loads(findSchema.get("properties")))
					reqSchema = ULTIPA_REQUEST.Schema(edge.schema, propertyList)
					batches[edge.schema].Schema = reqSchema
				else:
					if edge.schema is None:
						raise ParameterException(err=f"Row [{index}]:Please set schema name for edge.")
					else:
						raise ParameterException(err=f"Row [{index}]:Edge Schema not found {edge.schema}.")
			batches.get(edge.schema).Edges.append(edge)
		for key in batches:
			batch = batches.get(key)
			Result.update(
				{key: self.insertEdgesBatchBySchema(schema=batch.Schema, rows=batch.Edges, config=config)})

		newStatusMsg = ""
		newCode = None
		for i, key in enumerate(Result):
			ret = Result.get(key)
			newStatusMsg += f"{key}:{ret.status.message} "
			if ret.status.code != ULTIPA.Code.SUCCESS and not newCode:
				newCode = ret.status.code
		if newCode is None:
			newCode = ULTIPA.Code.SUCCESS
		status = ULTIPA_RESPONSE.Status(newCode, newStatusMsg)
		newResponse = ULTIPA_RESPONSE.ResponseBatchAutoInsert(status=status)
		newResponse.data = Result
		return newResponse

	def _InsertByCSV(self, csvPath: str, type: ULTIPA.DBType, config: ULTIPA_REQUEST.InsertConfig,
					 schemaName: str = None) -> ULTIPA_RESPONSE.ResponseBatchAutoInsert:
		rows = []
		propertyType = []
		properties = []
		types = []
		with open(csvPath, "r", encoding="utf-8-sig") as csvfile:
			reader = csv.reader(csvfile)
			for i, line in enumerate(reader):
				if i == 0:
					for i, property in enumerate(line):
						k1, k2 = property.split(":")
						propertyType.append({k1: k2})
						types.append({"index": i, "type": k2})
						properties.append(k1)
					continue
				for i in types:
					if i.get("type") in ["int", "int32", "int64"]:
						if line[i.get("index")] == "":
							line[i.get("index")] = 0
							continue
						line[i.get("index")] = int(line[i.get("index")])
					if i.get("type") in ["float", "double"]:
						if line[i.get("index")] == "":
							line[i.get("index")] = 0.0
							continue
						line[i.get("index")] = float(line[i.get("index")])
				line = dict(zip(properties, line))
				if i == 0:
					print(line.keys())
				if type == ULTIPA.DBType.DBNODE:
					if line.get("_uuid"):
						uuid = line.get("_uuid")
						line.__delitem__("_uuid")
						rows.append(ULTIPA.Node(line, schema_name=schemaName, uuid=int(uuid)))
					elif line.get("_id"):
						id = line.get("_id")
						line.__delitem__("_id")
						rows.append(ULTIPA.Node(line, schema_name=schemaName, id=id))
					else:
						rows.append(ULTIPA.Node(line, schema_name=schemaName))

				elif type == ULTIPA.DBType.DBEDGE:
					if line.get("_from_uuid") and line.get("_to_uuid"):
						from_uuid = line.get("_from_uuid")
						line.__delitem__("_from_uuid")
						to_uuid = line.get("_to_uuid")
						line.__delitem__("_to_uuid")
						line.__delitem__("_id")
						rows.append(
							ULTIPA.Edge(line, schema_name=schemaName, from_uuid=int(from_uuid),
										to_uuid=int(to_uuid)))
					elif line.get("_from_id") and line.get("_to_id"):
						from_id = line.get("_from_id")
						line.__delitem__("_from_id")
						to_id = line.get("_to_id")
						line.__delitem__("_to_id")
						if line.get("_id"):
							line.__delitem__("_id")
						rows.append(ULTIPA.Edge(line, schema_name=schemaName, from_id=from_id, to_id=to_id))

		if type == ULTIPA.DBType.DBNODE:
			return self.insertNodesBatchAuto(rows, config)
		else:
			return self.insertEdgesBatchAuto(rows, config)


# -*- coding: utf-8 -*-
# @Time    : 2023/1/16 11:06 上午
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : propertyUtils.py
from ultipa.types import ULTIPA
from typing import List


def isBasePropertyType(type: ULTIPA.CreatePropertyType):
	if type in [ULTIPA.CreatePropertyType.PROPERTY_STRING,
		ULTIPA.CreatePropertyType.PROPERTY_INT,
		ULTIPA.CreatePropertyType.PROPERTY_INT64,
		ULTIPA.CreatePropertyType.PROPERTY_UINT32,
		ULTIPA.CreatePropertyType.PROPERTY_UINT64,
		ULTIPA.CreatePropertyType.PROPERTY_FLOAT,
		ULTIPA.CreatePropertyType.PROPERTY_DOUBLE,
		ULTIPA.CreatePropertyType.PROPERTY_DATETIME,
		ULTIPA.CreatePropertyType.PROPERTY_TIMESTAMP,
		ULTIPA.CreatePropertyType.PROPERTY_TEXT]:
		return True
	return False

def propertyGet(type: ULTIPA.PropertyType):
	return type


def getPropertyTypesDesc(type:ULTIPA.CreatePropertyType,subTypes:List[ULTIPA.CreatePropertyType]):

	if type==ULTIPA.CreatePropertyType.PROPERTY_LIST:
		subType = subTypes[0]
		if isBasePropertyType(subType):
			return f"{subType}[]"


	return type



from ultipa.operate.base_extra import BaseExtra
from ultipa.utils import UQLMAKER, CommandList
from ultipa.types import ULTIPA_REQUEST, ULTIPA_RESPONSE
from ultipa.utils.errors import ParameterException
from ultipa.utils.format import FormatType
from ultipa.utils.ResposeFormat import ResponseKeyFormat

JSONSTRING_KEYS = ["graphPrivileges", "systemPrivileges", "policies", "policy", "privilege"]
formatdata = ['graph_privileges']


class UserExtra(BaseExtra):

	def GRPATH_PRIVILEGES_DATA_FORMAT(self, obj):
		if isinstance(obj.get('graph_privileges'), list):
			resr = FormatType.graphPrivileges(obj.get('graph_privileges'))
			return resr
		else:
			return '[]'

	def showUser(self,
				 requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.ResponseListUser:
		'''
		:EXP: conn.listUser()
		:return:
		'''
		uqlMaker = UQLMAKER(command=CommandList.showUser, commonParams=requestConfig)
		res = self.UqlListSimple(uqlMaker=uqlMaker,
								 responseKeyFormat=ResponseKeyFormat(jsonKeys=JSONSTRING_KEYS, dataFormat=formatdata))
		return res

	def getUser(self, request: ULTIPA_REQUEST.GetUser,
				requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.ResponseUser:
		'''
		:EXP: conn.getUser(ULTIPA_REQUEST.GetUser(username='Test'))
		:param request: ULTIPA_REQUEST
		:return:
		'''
		uqlMaker = UQLMAKER(command=CommandList.getUser, commonParams=requestConfig)
		uqlMaker.setCommandParams(request.username)
		res = self.UqlListSimple(uqlMaker=uqlMaker,
								 responseKeyFormat=ResponseKeyFormat(jsonKeys=JSONSTRING_KEYS, dataFormat=formatdata))
		if isinstance(res.data, list) and len(res.data) > 0:
			res.data = res.data[0]
		return res

	def getSelfInfo(self,
					requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.ResponseUser:
		'''
		:EXP: conn.getUser(ULTIPA_REQUEST.GetUser(username='Test'))
		:param request: ULTIPA_REQUEST
		:return:
		'''
		uqlMaker = UQLMAKER(command=CommandList.getSelfInfo, commonParams=requestConfig)
		res = self.UqlListSimple(uqlMaker=uqlMaker,
								 responseKeyFormat=ResponseKeyFormat(jsonKeys=JSONSTRING_KEYS, dataFormat=formatdata))

		return res

	def createUser(self, request: ULTIPA_REQUEST.CreateUser,
				   requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		:EXP: conn.createUser(ULTIPA_REQUEST.CreateUser(username="Test", password="Test"))
		:param request: ULTIPA_REQUEST
		:return:
		'''
		uqlMaker = UQLMAKER(command=CommandList.createUser, commonParams=requestConfig)
		params = []
		if request.username:
			params.append(request.username)
		else:
			raise ParameterException(err='username is a required parameter')

		if request.password:
			params.append(request.password)
		else:
			raise ParameterException(err='password is a required parameter')

		if request.graph_privileges:
			params.append(request.graph_privileges)
		else:
			params.append({})

		if request.system_privileges:
			params.append(request.system_privileges)
		else:
			params.append([])

		if request.policies:
			params.append(request.policies)
		else:
			params.append([])

		uqlMaker.setCommandParams(params)
		return self.uqlSingle(uqlMaker=uqlMaker)

	def dropUser(self, request: ULTIPA_REQUEST.DropUser,
				 requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		:EXP: conn.deleteUser(ULTIPA_REQUEST.DeleteUser(username=username))
		:param request: ULTIPA_REQUEST
		:return:
		'''
		uqlMaker = UQLMAKER(command=CommandList.dropUser, commonParams=requestConfig)
		uqlMaker.setCommandParams(request.username)
		return self.uqlSingle(uqlMaker=uqlMaker)

	def alterUser(self, request: ULTIPA_REQUEST.AlterUser,
				  requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		:EXP: conn.updateUser(ULTIPA_REQUEST.UpdateUser(username='Test',password='test',privileges=['DELETE','UPDATE'],policies=['ROOT']))
		:param request: ULTIPA_REQUEST
		:return:
		'''
		uqlMaker = UQLMAKER(command=CommandList.alterUser, commonParams=requestConfig)
		if request.username:
			uqlMaker.setCommandParams(request.username)
		else:
			raise ParameterException(err='username is a required parameter')

		paramsDict = {}
		if request.password:
			paramsDict.setdefault('password', request.password)

		if request.graph_privileges:
			paramsDict.setdefault('graph_privileges', request.graph_privileges)

		if request.system_privileges:
			paramsDict.setdefault('system_privileges', request.system_privileges)

		if request.policies:
			paramsDict.setdefault('policies', request.policies)

		uqlMaker.addParam('set', paramsDict)
		return self.uqlSingle(uqlMaker=uqlMaker)



# -*- coding: utf-8 -*-

import threading
import time
import queue


class TooManyConnections(Exception):
	'''当连接数过多时，抛出该异常'''


class Expired(Exception):
	'''当连接不可用时，抛出该异常'''


class UsageExceeded(Expired):
	'''连接的使用次数超出限制'''


class TtlExceeded(Expired):
	'''连接使用时间超出 ttl 指定的生命周期'''


class IdleExceeded(Expired):
	'''闲置时间超出 idle 指定的时间'''


class WrapperConnection(object):
	'''用于包装连接池中的数据库连接，以便处理生命周期逻辑'''

	def __init__(self, pool, connection):
		self.pool = pool
		self.connection = connection
		self.usage = 0
		self.last = self.created = time.time()

	def using(self):
		'''当连接被调用时使用该方法，使用次数 +1'''
		self.usage += 1
		self.last = time.time()
		return self

	def reset(self):
		'''重置连接包装状态'''
		self.usage = self.last = self.created = 0

	def __enter__(self):
		return self.connection

	def __exit__(self, *exc_info):
		self.pool.release(self)


class ConnectionPool(object):
	'''连接池类，可用于 pymysql/memcache/redis/... 等

	可通过如下方式调用：
		pool = ConnectionPool(create=redis.Redis)

	也可以通过 lambda 的方式指定 create 调用：
		pool = ConnectionPool(create=lambda: redis.Redis(host='127.0.0.1'))

	或者通过 functools.partial
		from functools import partial
		pool = ConnectionPool(create=partial(redis.Redis, host='127.0.0.1'))
	'''

	__wrappers = {}

	def __init__(self, create, close=None, max_size=3, max_usage=0, ttl=0, idle=120, block=True):
		'''初始化参数

			create:    必须是一个可 callback 的函数
					   must be a callback function, creates connection
			close:     optional callback to close connection
			max_size:  最大连接数，当为 0 的时候则没有限制，不建议设置为 0
					   the maximum number of connections, 0 is no limit (not recommended)
			max_usage: 连接可使用次数，达到该次数后，连接将被释放/关闭
					   the number of times the connection can be used,
					   after this number of times, the connection will be released/closed
			ttl:       连接可使用时间，单位(秒)，当连接使用达到指定时间后，
					   连接将被释放/关闭
					   connection use time limit (seconds), when reached, the connection
					   will be released/closed
			idle:      连接空闲时间，单位(秒)，当连接在闲置指定时间后，将被释放/关闭
					   connection idle time (seconds), when reached, the connection
					   will be released/closed
			block:     当连接数满的时候，是否阻塞等待连接被释放,
					   输入 False 则在连接池满时会抛出异常
					   When the number of connections is full,
					   whether to block and wait for the connection to be released.
					   Enter False to throw an exception when the connection pool is full
		'''
		if not hasattr(create, '__call__'):
			raise ValueError('"create" argument is not callable')

		if close is not None and not hasattr(close, '__call__'):
			raise ValueError('"close" argument is not callable')

		self._create = create
		self._close = close
		self._max_size = int(max_size)
		self._max_usage = int(max_usage)
		self._ttl = int(ttl)
		self._idle = int(idle)
		self._block = bool(block)
		self._lock = threading.Condition()
		self._pool = queue.Queue()
		self._size = 0
		if self._pool.qsize() == 0:
			while self._max_size:
				self._pool.put_nowait(self._wrapper(self._create()))
				self._max_size -= 1

	def item(self):
		'''可通过 with ... as ... 语法调用

			pool = ConnectionPool(create=redis.Redis)
			with pool.item() as redis:
				redis.set('foo', 'bar)
		'''
		self._lock.acquire()

		try:
			while (self._max_size and self._pool.empty() and self._size >= self._max_size):
				if not self._block:
					raise TooManyConnections('Too many connections')

				self._lock.wait()  # 等待闲置连接
			if self._pool.qsize() == 0:
				while self._max_size:
					self._pool.put_nowait(self._wrapper(self._create()))
					self._max_size -= 1
			try:
				wrapped = self._pool.get_nowait()  # 从空闲连接池中获取一个
				if self._idle and (wrapped.last + self._idle) < time.time():
					self._destroy(wrapped, isCreate=True)
					raise IdleExceeded('Idle exceeds %d secs' % self._idle)
			except (queue.Empty, IdleExceeded):
				wrapped = self._wrapper(self._create())  # 创建新连接
				self._size += 1
		finally:
			self._lock.release()

		return wrapped.using()

	def release(self, conn):
		'''释放一个连接，让连接重回到连接池中

		当连接使用超过限制/超过限定时间时，连接将被销毁
		'''
		self._lock.acquire()
		wrapped = self._wrapper(conn)

		try:
			self._test(wrapped)
		except Expired:
			self._destroy(wrapped)
		else:
			self._pool.put_nowait(wrapped)
			self._lock.notifyAll()  # 通知其它线程，有闲置连接可用
		finally:
			self._lock.release()

	def _destroy(self, wrapped, isCreate: bool = False):
		'''销毁一个连接'''
		if self._close:
			self._close(wrapped.connection)

		self._unwrapper(wrapped)
		self._size += 1
		if isCreate:
			self._pool.put_nowait(self._wrapper(self._create()))
			self._size -= 1

	def _wrapper(self, conn):
		'''利用 id 地址，对连接进行包装'''
		if isinstance(conn, WrapperConnection):
			return conn

		_id = id(conn)

		if _id not in self.__wrappers:
			self.__wrappers[_id] = WrapperConnection(self, conn)

		return self.__wrappers[_id]

	def _unwrapper(self, wrapped):
		'''取消对连接的包装'''
		if not isinstance(wrapped, WrapperConnection):
			return

		_id = id(wrapped.connection)
		wrapped.reset()
		del wrapped

		if _id in self.__wrappers:
			del self.__wrappers[_id]

	def _test(self, wrapped):
		'''测试连接的可用性，当不可用时，抛出 Expired 异常'''
		if self._max_usage and wrapped.usage >= self._max_usage:
			raise UsageExceeded('Usage exceeds %d times' % self._max_usage)

		if self._ttl and (wrapped.created + self._ttl) < time.time():
			raise TtlExceeded('TTL exceeds %d secs' % self._ttl)



# -*- coding: utf-8 -*-
# @Time    : 2023/7/18 09:21
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : clientInfo.py
from datetime import datetime
import grpc
import copy
import pytz
from tzlocal import get_localzone
from ultipa.proto import ultipa_pb2_grpc
from ultipa.utils.password2md5 import passwrod2md5


class ClientInfo:
	def __init__(self, Rpcsclient: ultipa_pb2_grpc.UltipaRpcsStub, Controlsclient: ultipa_pb2_grpc.UltipaControlsStub,
				 metadata: any, graphSetName: str, host: str):
		self.Rpcsclient = Rpcsclient
		self.Controlsclient = Controlsclient
		self.host = host
		self.metadata = metadata
		self.graphSetName = graphSetName


class GrpcClientInfo:
	Rpcsclient: ultipa_pb2_grpc.UltipaRpcsStub
	Controlsclient: ultipa_pb2_grpc.UltipaControlsStub
	host: str
	username: str
	password: str

	def __init__(self, host: str, username: str, password: str, crt: str, maxRecvSize: int = -1):
		self.host = host
		self._metadata = [('user', username), ('password', passwrod2md5(password))]
		if self.host.startswith("http"):
			channel = grpc.secure_channel(self.host.replace("https://","").replace("http://",""), grpc.ssl_channel_credentials(), options=(
				('grpc.max_send_message_length', -1), ('grpc.max_receive_message_length', maxRecvSize),
				("grpc.keepalive_timeout_ms", 1500)
			))
			self.Rpcsclient = ultipa_pb2_grpc.UltipaRpcsStub(channel=channel, )
			self.Controlsclient = ultipa_pb2_grpc.UltipaControlsStub(channel=channel, )
			return
		if crt:
			credentials = grpc.ssl_channel_credentials(root_certificates=crt)
			channel = grpc.secure_channel(self.host, credentials, options=(
				('grpc.ssl_target_name_override', 'ultipa'), ('grpc.default_authority', 'ultipa'),
				('grpc.max_send_message_length', -1), ('grpc.max_receive_message_length', maxRecvSize),
				("grpc.keepalive_timeout_ms", 1500)
			))
			self.Rpcsclient = ultipa_pb2_grpc.UltipaRpcsStub(channel=channel, )
			self.Controlsclient = ultipa_pb2_grpc.UltipaControlsStub(channel=channel, )
		else:
			channel = grpc.insecure_channel(self.host, options=(
				('grpc.max_send_message_length', -1), ('grpc.max_receive_message_length', maxRecvSize),
				("grpc.keepalive_timeout_ms", 1500)))
			self.Rpcsclient = ultipa_pb2_grpc.UltipaRpcsStub(channel=channel, )
			self.Controlsclient = ultipa_pb2_grpc.UltipaControlsStub(channel=channel, )

	def getMetadata(self, graphSetName, timeZone, timeZoneOffset):
		metadata = copy.deepcopy(self._metadata)
		metadata.append(('graph_name', graphSetName))
		if timeZone is None and timeZoneOffset is None:
			timeZone = get_localzone().__dict__.get("_key")
			tz = pytz.timezone(timeZone)
			timeZoneOffset = tz.utcoffset(datetime.now()).total_seconds()
			metadata.append(('tz_offset', str(timeZoneOffset)))
		if timeZone is not None and timeZoneOffset is None:
			metadata.append(('tz', timeZone))
		if timeZone is None and timeZoneOffset is not None:
			metadata.append(('tz_offset', str(timeZoneOffset)))
		return metadata



import hashlib
import os
from ultipa.operate.base_extra import BaseExtra
from ultipa.proto import ultipa_pb2
from ultipa.types import ULTIPA_REQUEST, ULTIPA, ULTIPA_RESPONSE
from ultipa.utils import UQLMAKER, CommandList, errors
from ultipa.utils.fileSize import read_in_chunks
from ultipa.utils.format import FormatType
from ultipa.utils.ResposeFormat import ResponseKeyFormat


class ALGO_RESULT:
	ALGO_RESULT_UNSET = -1
	WRITE_TO_FILE = 1
	WRITE_TO_DB = 2
	WRITE_TO_CLIENT = 4
	WRITE_TO_VISUALIZATION = 8


class AlgoExtra(BaseExtra):
	JSONSTRING_KEYS = ["param"]

	def showAlgo(self,
				 requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.ResponseListAlgo:
		uqlMaker = UQLMAKER(command=CommandList.showAlgo, commonParams=requestConfig)
		res = self.UqlListSimple(uqlMaker=uqlMaker,responseKeyFormat=ResponseKeyFormat(jsonKeys=self.JSONSTRING_KEYS))
		if res.status.code == ULTIPA.Code.SUCCESS:
			for algo in res.data:
				try:
					result_opt = int(algo.param.get("result_opt", 0))
				except Exception as e:
					raise errors.ParameterException(e)
				result_opt_obj = ULTIPA_RESPONSE.AlgoResultOpt()
				result_opt_obj.can_realtime = True if result_opt & ALGO_RESULT.WRITE_TO_CLIENT else False
				result_opt_obj.can_visualization = True if result_opt & ALGO_RESULT.WRITE_TO_VISUALIZATION else False
				result_opt_obj.can_write_back = True if result_opt & (
						ALGO_RESULT.WRITE_TO_DB | ALGO_RESULT.WRITE_TO_FILE) else False
				algo.__setattr__("result_opt", result_opt_obj)
		return res

	def __make_message(self, filename, md5, chunk):
		return ultipa_pb2.InstallAlgoRequest(
			file_name=filename, md5=md5, chunk=chunk
		)

	def __generate_messages(self, request: ULTIPA_REQUEST.InstallAlgo):
		messages = []
		file_object = open(request.soPath, 'rb')
		somd5 = hashlib.md5(file_object.read()).hexdigest()
		file_object.close()

		for chunk in read_in_chunks(request.soPath):
			filename = os.path.basename(request.soPath)
			messages.append(self.__make_message(filename, somd5, chunk))

		file_object = open(request.configPath, 'rb')
		configmd5 = hashlib.md5(file_object.read()).hexdigest()
		file_object.close()
		for chunk in read_in_chunks(request.configPath):
			filename = os.path.basename(request.configPath)
			messages.append(self.__make_message(filename, configmd5, chunk))
		for msg in messages:
			yield msg

	def installAlgo(self, request: ULTIPA_REQUEST.InstallAlgo,
					requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.Response:
		requestConfig.useMaster = True
		clientInfo = self.getClientInfo(graphSetName=requestConfig.graphName, useMaster=requestConfig.useMaster,isGlobal=True)
		response = ULTIPA_RESPONSE.Response()
		try:
			if os.path.exists(request.soPath) and os.path.exists(request.configPath):
				installRet = clientInfo.Controlsclient.InstallAlgo(self.__generate_messages(request),
																   metadata=clientInfo.metadata)
				status = FormatType.status(installRet.status)
				response.status = status
		except Exception as e:
			try:
				message = str(e._state.code) + ' : ' + str(e._state.details)
			except:
				message = str(e)

			response.status = ULTIPA.Status(code=ULTIPA.Code.UNKNOW_ERROR, message=message)

		if self.defaultConfig.responseWithRequestInfo:
			response.req = ULTIPA.ReturnReq(self.graphSetName, "InstallAlgo",
											requestConfig.useHost if requestConfig.useHost else self.host, requestConfig.retry,
											False)
		return response

	def uninstallAlgo(self, request: ULTIPA_REQUEST.UninstallAlgo,
					  requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.Response:
		requestConfig.useMaster=True
		clientInfo = self.getClientInfo(graphSetName=requestConfig.graphName, useMaster=requestConfig.useMaster,isGlobal=True)
		arequest = ultipa_pb2.UninstallAlgoRequest(algo_name=request.algoName)
		installRet = clientInfo.Controlsclient.UninstallAlgo(arequest, metadata=clientInfo.metadata)
		status = FormatType.status(installRet.status)
		response = ULTIPA_RESPONSE.Response(status=status)
		if self.defaultConfig.responseWithRequestInfo:
			response.req = ULTIPA.ReturnReq(self.graphSetName, "InstallAlgo",
											requestConfig.useHost if requestConfig.useHost else self.host, requestConfig.retry,
											False)
		return response

	def installExtaAlgo(self, request: ULTIPA_REQUEST.InstallExtaAlgo,
					requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.Response:
		requestConfig.useMaster = True
		clientInfo = self.getClientInfo(graphSetName=requestConfig.graphName, useMaster=requestConfig.useMaster)
		response = ULTIPA_RESPONSE.Response()
		try:
			if os.path.exists(request.soPath) and os.path.exists(request.configPath):
				installRet = clientInfo.Controlsclient.InstallExta(self.__generate_messages(request),
																   metadata=clientInfo.metadata)
				status = FormatType.status(installRet.status)
				response.status = status
		except Exception as e:
			try:
				message = str(e._state.code) + ' : ' + str(e._state.details)
			except:
				message = str(e)

			response.status = ULTIPA.Status(code=ULTIPA.Code.UNKNOW_ERROR, message=message)

		if self.defaultConfig.responseWithRequestInfo:
			response.req = ULTIPA.ReturnReq(self.graphSetName, "InstallAlgo",
											requestConfig.useHost if requestConfig.useHost else self.host,
											requestConfig.retry,
											False)
		return response

	def uninstallExtaAlgo(self, request: ULTIPA_REQUEST.UninstallExtaAlgo,
					  requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.Response:
		requestConfig.useMaster = True
		clientInfo = self.getClientInfo(graphSetName=requestConfig.graphName, useMaster=requestConfig.useMaster)
		arequest = ultipa_pb2.UninstallExtaRequest(exta_name=request.algoName)
		installRet = clientInfo.Controlsclient.UninstallExta(arequest, metadata=clientInfo.metadata)
		status = FormatType.status(installRet.status)
		response = ULTIPA_RESPONSE.Response(status=status)
		if self.defaultConfig.responseWithRequestInfo:
			response.req = ULTIPA.ReturnReq(self.graphSetName, "InstallAlgo",
											requestConfig.useHost if requestConfig.useHost else self.host,
											requestConfig.retry,
											False)
		return response






import os


def get_fileSize(FilePath:str):
    fsize = os.path.getsize(FilePath)
    fsize = fsize/float(1024*1024)
    return round(fsize,2)


def read_in_chunks(filePath, chunk_size=1024*1024):
    """
    Lazy function (generator) to read a file piece by piece.
    Default chunk size: 1M
    You can set your own chunk size
    """
    file_object = open(filePath,'rb')
    while True:
        chunk_data = file_object.read(chunk_size)
        if not chunk_data:
            file_object.close()
            break
        yield chunk_data

if __name__ == "__main__":
    filePath = '/Users/ultipa/work/ultipa-python-sdk2/algo/khop_all.yml'
    for chunk in read_in_chunks(filePath):
        print(type(chunk))


'''
自定义错误异常
'''
class ParameterException(Exception):
    def __init__(self,err='Parameter error!'):
        Exception.__init__(self,err)


class ServerException(Exception):
    def __init__(self,err='Server connection failed!'):
        Exception.__init__(self,err)


class SerializeException(Exception):
    def __init__(self,err='Serialize failed!'):
        Exception.__init__(self,err)



class SettingException(Exception):
    def __init__(self,err='Setting error!'):
        Exception.__init__(self,err)


def checkError(error: str):
    if "large" in error:
        return "argument out of range"
    else:
        return error


# -*-coding:utf-8-*-
import uuid
import json
from enum import Enum
from typing import List, Dict
from ultipa.proto import ultipa_pb2
from typing import TypeVar
from ultipa.utils.ResposeFormat import ResponseKeyFormat
from ultipa.utils.convert import convertToListAnyObject, convertTableToDict, convertToAnyObject, \
	convertAlgoList2AnyObject
from ultipa.utils.errors import ParameterException
from ultipa.utils.logger import LoggerConfig


# T = TypeVar('T')


class DBType:
	DBNODE = ultipa_pb2.DBNODE
	DBEDGE = ultipa_pb2.DBEDGE


class TruncateType:
	NODES = 'nodes'
	EDGES = 'edges'


class DirectionType:
	left = 'left'
	right = 'right'


class CreatePropertyType:
	PROPERTY_INT = 'int32'
	PROPERTY_STRING = 'string'
	PROPERTY_FLOAT = 'float'
	PROPERTY_DOUBLE = 'double'
	PROPERTY_UINT32 = 'uint32'
	PROPERTY_INT64 = 'int64'
	PROPERTY_UINT64 = 'uint64'
	PROPERTY_DATETIME = 'datetime'
	PROPERTY_TIMESTAMP = 'timestamp'
	PROPERTY_TEXT = 'text'
	PROPERTY_UNSET = "unset"
	PROPERTY_POINT = "point"
	PROPERTY_DECIMAL = "decimal"
	PROPERTY_LIST = "list"
	PROPERTY_SET = "set"
	PROPERTY_MAP = "map"


# PROPERTY_BLOB = 'BLOB'


class TaskStatus:
	TASK_WAITING = 0
	TASK_COMPUTING = 1
	TASK_WRITEBACKING = 2
	TASK_DONE = 3
	TASK_FAILED = 4
	TASK_STOP = 5


TaskStatusString = {
	TaskStatus.TASK_WAITING: "TASK_WAITING",
	TaskStatus.TASK_COMPUTING: "TASK_COMPUTING",
	TaskStatus.TASK_WRITEBACKING: "TASK_WRITEBACKING",
	TaskStatus.TASK_DONE: "TASK_DONE",
	TaskStatus.TASK_FAILED: "TASK_FAILED",
	TaskStatus.TASK_STOP: "TASK_STOP"
}


class InsertType:
	NORMAL = ultipa_pb2.NORMAL
	OVERWRITE = ultipa_pb2.OVERWRITE
	UPSERT = ultipa_pb2.UPSERT


class PropertyType:
	PROPERTY_UNSET = ultipa_pb2.UNSET
	PROPERTY_INT32 = ultipa_pb2.INT32
	PROPERTY_STRING = ultipa_pb2.STRING
	PROPERTY_FLOAT = ultipa_pb2.FLOAT
	PROPERTY_DOUBLE = ultipa_pb2.DOUBLE
	PROPERTY_UINT32 = ultipa_pb2.UINT32
	PROPERTY_INT64 = ultipa_pb2.INT64
	PROPERTY_UINT64 = ultipa_pb2.UINT64
	PROPERTY_DATETIME = ultipa_pb2.DATETIME
	PROPERTY_TIMESTAMP = ultipa_pb2.TIMESTAMP
	PROPERTY_TEXT = ultipa_pb2.TEXT
	PROPERTY_BLOB = ultipa_pb2.BLOB
	PROPERTY_POINT = ultipa_pb2.POINT
	PROPERTY_DECIMAL = ultipa_pb2.DECIMAL
	PROPERTY_LIST = ultipa_pb2.LIST
	PROPERTY_SET = ultipa_pb2.SET
	PROPERTY_MAP = ultipa_pb2.MAP
	PROPERTY_NULL = ultipa_pb2.NULL_
	PROPERTY_UUID = -1
	PROPERTY_ID = -2
	PROPERTY_FROM = -3
	PROPERTY_FROM_UUID = -4
	PROPERTY_TO = -5
	PROPERTY_TO_UUID = -6
	PROPERTY_IGNORE = -7


class ResultType:
	RESULT_TYPE_UNSET = ultipa_pb2.RESULT_TYPE_UNSET
	RESULT_TYPE_PATH = ultipa_pb2.RESULT_TYPE_PATH
	RESULT_TYPE_NODE = ultipa_pb2.RESULT_TYPE_NODE
	RESULT_TYPE_EDGE = ultipa_pb2.RESULT_TYPE_EDGE
	RESULT_TYPE_ATTR = ultipa_pb2.RESULT_TYPE_ATTR
	# RESULT_TYPE_ARRAY = ultipa_pb2.RESULT_TYPE_ARRAY
	RESULT_TYPE_TABLE = ultipa_pb2.RESULT_TYPE_TABLE
	RESULT_TYPE_ExplainPlan = "ExplainPlan"

	@staticmethod
	def getTypeStr(type):
		if type == ResultType.RESULT_TYPE_PATH:
			return 'PATH'
		elif type == ResultType.RESULT_TYPE_NODE:
			return 'NODE'
		elif type == ResultType.RESULT_TYPE_EDGE:
			return "EDGE"
		elif type == ResultType.RESULT_TYPE_ATTR:
			return "ATTR"
		elif type == PropertyType.PROPERTY_LIST:
			return "LIST"
		elif type == ResultType.RESULT_TYPE_TABLE:
			return "TABLE"
		elif type == ResultType.RESULT_TYPE_UNSET:
			return "UNSET"
		elif type == ResultType.RESULT_TYPE_ExplainPlan:
			return "EXPLAINPLAN"
		else:
			return type


class Code:
	SUCCESS = ultipa_pb2.SUCCESS
	FAILED = ultipa_pb2.FAILED
	PARAM_ERROR = ultipa_pb2.PARAM_ERROR
	BASE_DB_ERROR = ultipa_pb2.BASE_DB_ERROR
	ENGINE_ERROR = ultipa_pb2.ENGINE_ERROR
	SYSTEM_ERROR = ultipa_pb2.SYSTEM_ERROR
	RAFT_REDIRECT = ultipa_pb2.RAFT_REDIRECT
	RAFT_LEADER_NOT_YET_ELECTED = ultipa_pb2.RAFT_LEADER_NOT_YET_ELECTED
	RAFT_LOG_ERROR = ultipa_pb2.RAFT_LOG_ERROR
	UQL_ERROR = ultipa_pb2.UQL_ERROR
	NOT_RAFT_MODE = ultipa_pb2.NOT_RAFT_MODE
	RAFT_NO_AVAILABLE_FOLLOWERS = ultipa_pb2.RAFT_NO_AVAILABLE_FOLLOWERS
	RAFT_NO_AVAILABLE_ALGO_SERVERS = ultipa_pb2.RAFT_NO_AVAILABLE_ALGO_SERVERS
	PERMISSION_DENIED = ultipa_pb2.PERMISSION_DENIED

	UNKNOW_ERROR = 1000


class FollowerRole:
	ROLE_UNSET = ultipa_pb2.ROLE_UNSET
	ROLE_READABLE = ultipa_pb2.ROLE_READABLE
	ROLE_ALGO_EXECUTABLE = ultipa_pb2.ROLE_ALGO_EXECUTABLE


class ServerStatus:
	DEAD = ultipa_pb2.DEAD
	ALIVE = ultipa_pb2.ALIVE


class SchemaProperty:
	'''
	Schema properties
	'''

	name: str
	type: str
	description: str
	lte: str


class Schema:

	def __init__(self, description: str, schemaName: str, properties: List[SchemaProperty], total: int, type: str):
		self.description = description
		self.properties = properties
		self.name = schemaName
		self.total = total
		self.type = type

class RaftPeerInfo:
	def __init__(self, host, status=None, isLeader=False, isAlgoExecutable=False, isFollowerReadable=False,
				 isUnset=False):
		self.host = host
		self.status = status
		self.isLeader = isLeader
		self.isAlgoExecutable = isAlgoExecutable
		self.isFollowerReadable = isFollowerReadable
		self.isUnset = isUnset


class ClusterInfo:
	def __init__(self, redirect: str, raftPeers: List[RaftPeerInfo], leader: RaftPeerInfo = None):
		self.redirect = redirect
		self.leader = leader
		self.raftPeers = raftPeers


class Status:
	def __init__(self, code: Code, message: str, clusterInfo: ClusterInfo = None):
		self.code = code
		self.message = message.strip('\n')
		if clusterInfo:
			self.clusterInfo = clusterInfo


class BaseModel:

	def toJSON(self, pretty=False):
		try:
			if pretty:
				return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4, ensure_ascii=False)
			else:
				return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, ensure_ascii=False)
		except Exception as e:
			return self

	def toDict(self) -> Dict:
		return json.loads(self.toJSON())

	def __str__(self):
		return str(self.__dict__)


class Schema(BaseModel):

	def __init__(self, description: str, schemaName: str, properties: List[SchemaProperty], total: int, type: str):
		self.description = description
		self.properties = properties
		self.name = schemaName
		self.total = total
		self.type = type


class Property(BaseModel):
	lte: bool
	name: str
	type: str
	description: str
	schema: str


class Graph(BaseModel):
	id: str
	name: str
	totalNodes: str
	totalEdges: str
	description: str
	status: str


class Algo(BaseModel):
	name: str
	description: str
	version: str
	result_opt: str
	parameters: dict
	write_to_stats_parameters: dict
	write_to_db_parameters: dict
	write_to_file_parameters: dict


class Node(BaseModel):
	_index = None

	def __init__(self, values: Dict, schema: str = None, id: str = None, uuid: int = None, **kwargs):
		self.id = id
		if schema is None:
			if kwargs.get("schema_name") is not None:
				self.schema = kwargs.get("schema_name")
			else:
				self.schema = None
		else:
			self.schema = schema
		self.values = values
		self.uuid = uuid

	def getID(self):
		return self.id

	def getUUID(self):
		return self.uuid

	def getValues(self):
		return self.values

	def getSchema(self):
		return self.schema

	def get(self, propertyName: str):
		return self.values.get(propertyName)

	def set(self, propertyName: str, value):
		self.values.update({propertyName: value})

	def _getIndex(self):
		return self._index


class Edge(BaseModel):
	_index = None

	def __init__(self, values: Dict, from_id: str = None, from_uuid: int = None, to_id: str = None, to_uuid: int = None,
				 schema: str = None,
				 uuid: int = None, **kwargs):
		if schema is None:
			if kwargs.get("schema_name") is not None:
				self.schema = kwargs.get("schema_name")
			else:
				self.schema = None
		else:
			self.schema = schema

		self.from_id = from_id
		self.from_uuid = from_uuid
		self.to_id = to_id
		self.to_uuid = to_uuid
		self.values = values
		self.uuid = uuid

	def getUUID(self):
		return self.uuid

	def getFrom(self):
		return self.from_id

	def getTo(self):
		return self.to_id

	def getFromUUID(self):
		return self.from_uuid

	def getToUUID(self):
		return self.to_uuid

	def getValues(self):
		return self.values

	def getSchema(self):
		return self.schema

	def get(self, propertyName: str):
		return self.values.get(propertyName)

	def set(self, propertyName: str, value):
		self.values.update({propertyName: value})

	def _getIndex(self):
		return self._index


class EntityRow:
	_index = None

	def __init__(self, values: Dict, schema: str = None, id: str = None, from_id: str = None, to_id: str = None,
				 uuid: int = None, from_uuid: int = None, to_uuid: int = None, **kwargs):
		self.uuid = uuid
		self.id = id
		self.from_uuid = from_uuid
		self.to_uuid = to_uuid
		self.from_id = from_id
		self.to_id = to_id
		if schema is None:
			if kwargs.get("schema_name") is not None:
				self.schema = kwargs.get("schema_name")
			else:
				self.schema = None
		else:
			self.schema = schema
		self.values = values

	def _getIndex(self):
		return self._index


class Path(BaseModel):
	nodeSchemas: Dict[str, Schema] = {}
	edgeSchemas: Dict[str, Schema] = {}

	def __init__(self, nodes: List[Node], edges: List[Edge], nodeSchemas, edgeSchemas):
		self.nodes = nodes
		self.edges = edges
		self.nodeSchemas = nodeSchemas
		self.edgeSchemas = edgeSchemas


	def length(self):
		return len(self.edges)

	def getNodes(self):
		return self.nodes

	def getEdges(self):
		return self.edges


class PathAlias:
	def __init__(self, alias: str, paths: List[Path] = None):
		self.alias = alias
		if paths is None:
			paths = []
		self.paths = paths

	def length(self):
		return len(self.paths)

	def getNodes(self):
		nodes = [i.nodes for i in self.paths]
		return nodes

	def getEdges(self):
		edges = [i.edges for i in self.paths]
		return edges


class NodeAlias:
	def __init__(self, alias: str, nodes: List[Node] = None):
		self.alias = alias
		if nodes is None:
			nodes = []
		self.nodes = nodes


class Header:
	property_type: str
	property_name: str


class SchemaHeader:
	schema_name: str
	headers: List[Header]


class NodeTable:
	def __init__(self, schemas: List[object], nodeRows: List[Node] = None):
		self.schemas = schemas
		if nodeRows is None:
			nodeRows = []
		self.nodeRows = nodeRows

	def __del__(self):
		pass


class NodeEntityTable:
	def __init__(self, schemas: List[object], nodeRows: List[EntityRow] = None):
		self.schemas = schemas
		if nodeRows is None:
			nodeRows = []
		self.nodeRows = nodeRows

	def __del__(self):
		pass


class EdgeTable:
	def __init__(self, schemas: List[object], edgeRows: List[Edge] = None):
		self.schemas = schemas
		if edgeRows is None:
			edgeRows = []
		self.edgeRows = edgeRows


class EdgeEntityTable:
	def __init__(self, schemas: List[object], edgeRows: List[EntityRow] = None):
		self.schemas = schemas
		if edgeRows is None:
			edgeRows = []
		self.edgeRows = edgeRows

	def __del__(self):
		pass


class EntityTable:
	def __init__(self, schemas: List[object], entity_rows: List[EntityRow] = None):
		self.schemas = schemas
		if entity_rows is None:
			entity_rows = []
		self.entity_rows = entity_rows

	def __del__(self):
		pass


class EdgeAlias:
	def __init__(self, alias: str, edges: List[Edge]):
		self.alias = alias
		self.edges = edges





class Attr:
	def __init__(self, alias: str, values: any, type: ResultType = None,type_desc:str=None):
		self.name = alias
		self.values = values
		self.type = type
		self.type_desc = type_desc

class AttrNode:
	def __init__(self, alias: str, values:  List[List[Node]], type: ResultType = None,type_desc:str=None):
		self.name = alias
		self.values = values
		self.type = type
		self.type_desc = type_desc

class AttrEdge:
	def __init__(self, alias: str, values: List[List[Edge]], type: ResultType = None,type_desc:str=None):
		self.name = alias
		self.values = values
		self.type = type
		self.type_desc = type_desc

class AttrPath:
	def __init__(self, alias: str, values: List[List[Path]], type: ResultType = None,type_desc:str=None):
		self.name = alias
		self.values = values
		self.type = type
		self.type_desc = type_desc

class UltipaAttr:

	def __init__(self, type: PropertyType, values: any,has_attr_data:bool=False,has_ultipa_data:bool=False,type_desc: any = None):
		self.values = values
		self.type = type
		self.type_desc = type_desc
		self.has_attr_data = has_attr_data
		self.has_ultipa_data = has_ultipa_data


class AttrNewAlias:
	def __init__(self, alias: str, attr: UltipaAttr):
		self.alias = alias
		self.attr = attr



class ResultAlias:
	def __init__(self, alias: str, result_type: int):
		self.alias = alias
		self.result_type = result_type


class Table(BaseModel):
	def __init__(self, table_name: str, headers: List[dict], table_rows: List[List]):
		self.name = table_name
		self.rows = table_rows
		self.headers = headers

	def getHeaders(self):
		return self.headers

	def getRows(self):
		return self.rows

	def getName(self):
		return self.name


class Values:
	def __init__(self, key: str, value: str):
		self.key = key
		self.value = value


class ArrayAlias:
	def __init__(self, alias: str, elements):
		self.alias = alias
		self.elements = elements


class ExplainPlan:
	def __init__(self, alias, childrenNum, uql, infos):
		self.alias = alias
		self.children_num = childrenNum
		self.uql = uql
		self.infos = infos
		self.id = str(uuid.uuid4())


class DataItem(BaseModel):

	def __init__(self, alias: str, data: any, type: str):
		self.alias = alias
		self.data = data
		self.type = type

	def asNodes(self) -> List[Node]:
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			if self.data is None:
				return []
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_NODE):
			error = f"DataItem {self.alias} is not Type Node"
			raise ParameterException(error)
		return self.data

	def asFirstNodes(self) -> Node:
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_NODE):
			error = f"DataItem {self.alias} is not Type Node"
			raise ParameterException(error)
		return self.data[0] if len(self.data) > 0 else None

	def asEdges(self) -> List[Edge]:
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			if self.data is None:
				return []
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_EDGE):
			error = f"DataItem {self.alias} is not Type Edge"
			raise ParameterException(error)
		return self.data

	def asFirstEdges(self) -> Edge:
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_EDGE):
			error = f"DataItem {self.alias} is not Type Edge"
			raise ParameterException(error)
		return self.data[0] if len(self.data) > 0 else None

	def asPaths(self) -> List[Path]:
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			if self.data is None:
				return []
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_PATH):
			error = f"DataItem {self.alias} is not Type Path"
			raise ParameterException(error)
		return self.data

	def asAttr(self) -> Attr:
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_ATTR):
			error = f"DataItem {self.alias} is not Type Attribute list"
			raise ParameterException(error)

		return self.data

	def asNodeList(self) -> AttrNode:
		return self.asAttr()

	def asEdgeList(self) -> AttrEdge:
		return self.asAttr()

	def asPathList(self) -> AttrPath:
		return self.asAttr()


	def asTable(self) -> Table:
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
			error = f"DataItem {self.alias} is not Type Table"
			raise ParameterException(error)
		return self.data

	def asSchemas(self) -> List[Schema]:
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			if self.data is None:
				return []
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
			error = f"DataItem {self.alias} is not Type Table"
			raise ParameterException(error)
		alias = self.data.getName()
		headers = self.data.getHeaders()
		rows = self.data.getRows()
		tableListDict = convertTableToDict(rows, headers)
		REPLACE_KEYS = {
			"totalNodes": "total",
			"totalEdges": "total",
		}
		BOOL_KEYS = ["index", "lte"]
		JSON_KEYS = ["properties"]
		convert2Int = ["totalNodes", "totalEdges"]
		responseKeyFormat = ResponseKeyFormat(keyReplace=REPLACE_KEYS, boolKeys=BOOL_KEYS, jsonKeys=JSON_KEYS,
											  convert2Int=convert2Int)
		if alias == "_nodeSchema":
			for data in tableListDict:
				data.update({"type": "Node"})
		elif alias == "_edgeSchema":
			for data in tableListDict:
				data.update({"type": "Edge"})
		data = responseKeyFormat.changeKeyValue(tableListDict)
		data = convertToListAnyObject(data)
		return data

	def asProperties(self) -> List[Property]:
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			if self.data is None:
				return []
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
			error = f"DataItem {self.alias} is not Type Table"
			raise ParameterException(error)
		headers = self.data.getHeaders()
		rows = self.data.getRows()
		table_rows_dict = convertTableToDict(rows, headers)
		BOOL_KEYS = ["lte"]
		responseKeyFormat = ResponseKeyFormat(boolKeys=BOOL_KEYS)
		data = responseKeyFormat.changeKeyValue(table_rows_dict)
		data = convertToListAnyObject(data)
		return data

	def asGraphs(self) -> List[Graph]:
		REPLACE_KEYS = {
			"graph": "name",
		}
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			if self.data is None:
				return []
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
			error = f"DataItem {self.alias} is not Type Table"
			raise ParameterException(error)
		headers = self.data.getHeaders()
		rows = self.data.getRows()
		table_rows_dict = convertTableToDict(rows, headers)
		responseKeyFormat = ResponseKeyFormat(keyReplace=REPLACE_KEYS)
		data = responseKeyFormat.changeKeyValue(table_rows_dict)
		data = convertToListAnyObject(data)
		return data

	def asAlgos(self) -> List[Algo]:
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			if self.data is None:
				return []
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
			error = f"DataItem {self.alias} is not Type Table"
			raise ParameterException(error)
		headers = self.data.getHeaders()
		rows = self.data.getRows()
		table_rows_dict = convertTableToDict(rows, headers)
		algo_list = []
		for data in table_rows_dict:
			algo_list.append(json.loads(data.get("param")))

		responseKeyFormat = ResponseKeyFormat()
		data = responseKeyFormat.changeKeyValue(algo_list)
		data = convertAlgoList2AnyObject(data)

		return data


	def asAny(self) -> any:
		return self.data

	def asKV(self):
		return self.toDict()


class BaseUqlReply:
	def __init__(self, paths: List[PathAlias], nodes: List[NodeAlias], edges: List[EdgeAlias], tables: List[Table],
				 attrs: List = None, resultAlias: List = None,
				 explainPlan: List[ExplainPlan] = None):
		self.paths = paths
		self.nodes = nodes
		self.edges = edges
		self.tables = tables
		self.attrs = attrs
		# self.arrays = arrays
		self.resultAlias = resultAlias
		self.explainPlan = explainPlan


class UltipaStatistics(BaseModel):
	def __init__(self, edge_affected: int, node_affected: int, engine_time_cost: int, total_time_cost: int):
		self.edgeAffected = edge_affected
		self.nodeAffected = node_affected
		self.engineCost = engine_time_cost
		self.totalCost = total_time_cost


class UqlReply(BaseModel):
	datas: List[DataItem]

	def __init__(self, dataBase: BaseUqlReply, aliasMap: dict = None, datas: List = None):
		if aliasMap == None:
			aliasMap = {}
		self._aliasMap = aliasMap
		if datas is None:
			datas = []
		self.datas: List[DataItem] = datas
		self.explainPlan: List[ExplainPlan] = []
		self._dataBase = dataBase

		for data in self._dataBase.paths:
			if self._aliasMap.get(data.alias):
				self._aliasMap[data.alias].data.extend(data.paths)
				continue
			self._aliasMap[data.alias] = DataItem(data.alias, data.paths,
												  ResultType.getTypeStr(ResultType.RESULT_TYPE_PATH))

		for data in self._dataBase.nodes:
			if self._aliasMap.get(data.alias):
				self._aliasMap[data.alias].data.extend(data.nodes)
				continue
			self._aliasMap[data.alias] = DataItem(data.alias, data.nodes,
												  ResultType.getTypeStr(ResultType.RESULT_TYPE_NODE))

		for data in self._dataBase.edges:
			if self._aliasMap.get(data.alias):
				self._aliasMap[data.alias].data.extend(data.edges)
				continue
			self._aliasMap[data.alias] = DataItem(data.alias, data.edges,
												  ResultType.getTypeStr(ResultType.RESULT_TYPE_EDGE))

		for data in self._dataBase.attrs:
			if self._aliasMap.get(data.name):
				self._aliasMap[data.name].data.append(data)
				continue
			self._aliasMap[data.name] = DataItem(data.name, data, ResultType.getTypeStr(ResultType.RESULT_TYPE_ATTR))

		for data in self._dataBase.tables:
			if self._aliasMap.get(data.name):
				self._aliasMap[data.name].data.extend(data)
				continue
			self._aliasMap[data.name] = DataItem(data.name, data, ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE))

		for data in self._dataBase.explainPlan:
			self.explainPlan.append(data)

		for data in self._dataBase.resultAlias:
			if self._aliasMap.get(data.alias):
				self.datas.append(self._aliasMap[data.alias])
		if not self.datas:
			for key in self._aliasMap:
				self.datas.append(self._aliasMap[key])



class Retry:
	def __init__(self, canRetry, currentRetry, nextRetry):
		self.canRetry = canRetry
		self.currentRetry = currentRetry
		self.nextRetry = nextRetry


class ReturnReq:
	def __init__(self, graphSetName: str, uql: str, host: str, retry: Retry, uqlIsExtra: bool):
		self.graph_name = graphSetName
		self.uql = uql
		self.host = host
		self.Retry = retry
		self.uqlIsExtra = uqlIsExtra

class ExportReply:
	def __init__(self, data: List[NodeAlias]):
		self.data = data


class UltipaConfig:
	hosts: List[str] = []
	defaultGraph: str = 'default'
	timeoutWithSeconds: int = 3600
	responseWithRequestInfo: bool = False
	# 读一致性,如果为False 负载取节点执行
	consistency: bool = False
	uqlLoggerConfig: LoggerConfig = None
	heartBeat: int = 10
	maxRecvSize: int
	Debug: bool = False
	timeZone = None
	timeZoneOffset = None

	def __init__(self, hosts=None, username=None, password=None, crtFilePath=None, defaultGraph: str = defaultGraph,
				 timeout: int = timeoutWithSeconds, responseWithRequestInfo: bool = responseWithRequestInfo,
				 consistency: bool = consistency, heartBeat: int = 10, maxRecvSize: int = -1,
				 uqlLoggerConfig: LoggerConfig = uqlLoggerConfig, debug: bool = False, timeZone=None,
				 timeZoneOffset=None, **kwargs):
		if hosts is None:
			hosts = []
		self.hosts = hosts
		self.username = username
		self.password = password
		self.crtFilePath = crtFilePath
		self.defaultGraph = defaultGraph
		self.timeoutWithSeconds = timeout
		self.responseWithRequestInfo = responseWithRequestInfo
		self.consistency = consistency
		self.uqlLoggerConfig = uqlLoggerConfig
		self.heartBeat = heartBeat
		self.maxRecvSize = maxRecvSize
		self.Debug = debug
		self.timeZone = timeZone
		self.timeZoneOffset = timeZoneOffset
		if kwargs.get("timeoutWithSeconds") is not None:
			self.timeoutWithSeconds = kwargs.get("timeoutWithSeconds")

	def setDefaultGraphName(self, graph: str):
		self.defaultGraph = graph


class PaserAttrListData:
	def __init__(self, type, nodes: List[Node]=None, edges: List[Edge]=None, paths: List[Path]=None, attrs: UltipaAttr=None):
		self.type = type
		self.nodes = nodes
		self.edges = edges
		self.paths = paths
		self.attrs = attrs


class OrderType(Enum):
	desc = 'desc'
	asc = 'asc'


class UltipaEquation(Enum):
	sum = 'sum'
	count = 'count'



# -*- coding: utf-8 -*-
import datetime

from dateutil.parser import parse


def is_valid_date(strdate):
    '''判断是否是一个有效的日期字符串'''
    try:
        dt = parse(strdate)
    except Exception as e:
        return False



from ultipa.operate.base_extra import BaseExtra
from ultipa.types import ULTIPA_REQUEST, ULTIPA, ULTIPA_RESPONSE
from ultipa.utils import UQLMAKER, CommandList


class NodeExtra(BaseExtra):

	def searchNode(self, request: ULTIPA_REQUEST.SearchNode,
				   requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.ResponseSearchNode:
		'''
		:param requestConfig:
		:EXP: conn.searchNode(ULTIPA_REQUEST.SearchMate(id='1',select=['name']))
		:param request: ULTIPA_REQUEST
		:return:
		'''
		uqlMaker = UQLMAKER(command=CommandList.nodes, commonParams=requestConfig)
		if request.id:
			uqlMaker.setCommandParams(request.id)
		elif request.filter:
			uqlMaker.setCommandParams(request.filter)
		uqlMaker.addParam('as', request.select.aliasName)
		uqlMaker.addParam("return", request.select)
		res = self.uqlSingle(uqlMaker)
		if res.status.code != ULTIPA.Code.SUCCESS:
			return res

		return res

	def insertNode(self, request: ULTIPA_REQUEST.InsertNode,
				   requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.ResponseInsert:
		'''
		:param requestConfig:
		:EXP: conn.insertNode(ULTIPA_REQUEST.InsertNode(nodes=[{'name':'test'},{'name':'test2'}]))
		:param request: ULTIPA_REQUEST
		:return:
		'''
		uqlMaker = UQLMAKER(command=CommandList.insert, commonParams=requestConfig)
		if request.upsert:
			uqlMaker = UQLMAKER(command=CommandList.upsert, commonParams=requestConfig)
		if request.overwrite:
			uqlMaker.addParam('overwrite', "", required=False)
		if request.schemaName:
			uqlMaker.addParam('into', request.schemaName, required=False)

		uqlMaker.addParam('nodes', request.nodes)

		if request.isReturnID:
			uqlMaker.addParam('as', "nodes")
			uqlMaker.addParam('return', "nodes._uuid")

		res = self.uqlSingle(uqlMaker)
		if res.status.code != ULTIPA.Code.SUCCESS:
			return res
		if request.isReturnID:
			if len(res.aliases) > 0:
				res.data = res.items.get(res.aliases[0].alias).data
		return res

	def updateNode(self, request: ULTIPA_REQUEST.UpdateNode,
				   requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		:EXP: conn.updateNode(ULTIPA_REQUEST.UpdateMate(id='1',params={'age':30}))
		:param request: ULTIPA_REQUEST
		:param requestConfig:
		:return: status
		'''
		uqlMaker = UQLMAKER(command=CommandList.updateNodes, commonParams=requestConfig)
		if request.id:
			uqlMaker.setCommandParams(request.id)
		elif request.filter:
			uqlMaker.setCommandParams(request.filter)
		uqlMaker.addParam("set", request.values)
		uqlMaker.addParam("silent", request.silent)
		res = self.uqlSingle(uqlMaker)
		return res

	def deleteNode(self, request: ULTIPA_REQUEST.DeleteNode,
				   requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		:param requestConfig:
		:EXP: conn.deleteNode(ULTIPA_REQUEST.DeleteMate(filter={'name':'test'}))
		:param request: ULTIPA_REQUEST
		:return: status
		'''
		uqlMaker = UQLMAKER(command=CommandList.deleteNodes, commonParams=requestConfig)
		if request.id:
			uqlMaker.setCommandParams(request.id)
		elif request.filter:
			uqlMaker.setCommandParams(request.filter)
		res = self.uqlSingle(uqlMaker)
		return res



from ultipa.operate.base_extra import BaseExtra
from ultipa.types import ULTIPA, ULTIPA_REQUEST, ULTIPA_RESPONSE
from ultipa.utils import UQLMAKER, CommandList

BOOL_KEYS = ["index", "lte"]
REPLACE_KEYS = {
	"name": "schemaName",
	"type": "propertyType",
}


class SchemaExtra(BaseExtra):

	def createSchema(self, request: ULTIPA_REQUEST.CreateSchema,
					 requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()):

		'''
		// create node schema
		create().node_schema("<name>", "<description>");

		// create edge schema
		create().edge_schema("<name>", "<description>");
		:return:
		'''

		command = request.type == ULTIPA.DBType.DBNODE and CommandList.createNodeSchema or CommandList.createEdgeSchema
		commandP = [request.schemaName]
		if request.description:
			commandP.append(request.description)
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		uqlMaker.setCommandParams(commandP=commandP)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res

	def listSchema(self, request: ULTIPA_REQUEST.ShowSchema = None,
				   requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		return self.showSchema(request, requestConfig)

	def showSchema(self, request: ULTIPA_REQUEST.ShowSchema = None,
				   requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		// list all schema
		list().schema()

		// list node schema
		list().node_schema()

		// list edge schema()
		list().edge_schema()

		'''
		if request:
			if request.schemaType == ULTIPA.DBType.DBNODE:
				command = CommandList.showNodeSchema
			elif request.schemaType == ULTIPA.DBType.DBEDGE:
				command = CommandList.showEdgeSchema
			else:
				command = CommandList.showSchema

			if request.schemaName:
				commandP = '@' + request.schemaName
			else:
				commandP = ''
		else:
			command = CommandList.showSchema
			commandP = ''

		uqlMaker = UQLMAKER(command=command, commandP=commandP, commonParams=requestConfig)
		res = self.uqlSingle(uqlMaker)
		return res

	def alterSchema(self, request: ULTIPA_REQUEST.AlterSchema,
					requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()):
		'''
		// alter a node schema
		alter().node_schema(@<schema>).set({name:"<name>", description:"<desc>"})

		// alter an edge schema
		alter().edge_schema(@<schema>).set({name:"<name>", description:"<desc>"})
		:return:
		'''
		command = request.type == ULTIPA.DBType.DBNODE and CommandList.alterNodeSchema or CommandList.alterEdgeSchema
		commandP = '@' + request.schemaName
		update_dict = {}
		if request.new_schemaName:
			update_dict.setdefault('name', request.new_schemaName)
		if request.description:
			update_dict.update({'description': request.description})
		uqlMaker = UQLMAKER(command=command, commandP=commandP, commonParams=requestConfig)
		uqlMaker.addParam("set", update_dict)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res

	def dropSchema(self, request: ULTIPA_REQUEST.DropSchema,
				   requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()):

		'''
		// drop node schema
		drop().node_schema(@<schema>)

		// drop edge schema
		drop().edge_schema(@<schema>)
		:return:
		'''
		command = request.type == ULTIPA.DBType.DBNODE and CommandList.dropNodeSchema or CommandList.dropEdgeSchema
		commandP = '@' + request.schemaName

		uqlMaker = UQLMAKER(command=command, commandP=commandP, commonParams=requestConfig)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res



from ultipa.operate.base_extra import BaseExtra
from ultipa.types import ULTIPA_REQUEST, ULTIPA, ULTIPA_RESPONSE
from ultipa.utils import UQLMAKER, CommandList
import json
from typing import List


class TemplateExtra(BaseExtra):
	def template(self, request: ULTIPA_REQUEST.Template,
				 requestConfig: ULTIPA_REQUEST.RequestConfig = ULTIPA_REQUEST.RequestConfig()) -> ULTIPA_RESPONSE.Response:
		'''
		:EXP: conn.searchNode(ULTIPA_REQUEST.SearchMate(id='1',select=['name']))
		:param request: ULTIPA_REQUEST
		:return:
		'''
		command = request.isKhopTemplate and CommandList.khop or CommandList.template
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig, commandP=request.alias)
		uqlMaker.addTemplateParams(request.items)
		uqlMaker.addParam("limit", request.limit)
		uqlMaker.addParam("order_by", request.order_by)
		uqlMaker.addParam("return", request._return)
		if request.select:
			uqlMaker.addParam("select", request.select)
		res = self.uqlSingle(uqlMaker)
		if res.status.code != ULTIPA.Code.SUCCESS:
			return res
		return res
